/*
 *               ParaStation
 *
 * Copyright (C) 2005 Cluster Competence Center GmbH, Munich
 *
 * $Id: cscrambler.c 4292 2006-08-23 12:30:05Z hauke $
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <math.h>
#include <popt.h>

FILE *inp;
FILE *out;
FILE *left_over_file = NULL;

int arg_verbose = 0; 
int arg_nokeywords = 0;
int arg_savelongcomment = 0;
int arg_linerefs = 0;
int arg_simpleuniq = 0;
char *arg_infile = NULL;
char *arg_outfile = NULL;
char *arg_subs_file = NULL;
char *arg_subs_left_file = NULL;

#define D if (0)
//#define D
#define MAX_IDENT_LEN 128
char cur[3];
int current_line = 1;

void parse_opt(int argc, char **argv)
{
    const char *str;
    int    c;            /* used for argument parsing */
    poptContext optCon;   /* context for parsing command-line options */
    
    struct poptOption optionsTable[] = {
	{ "subs"  , 's', POPT_ARG_STRING, &arg_subs_file , 0,
	  "substitution file", "filename" },
	{ "leftover"  , 'l', POPT_ARG_STRING, &arg_subs_left_file , 0,
	  "left over file", "filename" },
	{ "output"  , 'o', POPT_ARG_STRING, &arg_outfile, 0,
	  "output file", "filename" },
	{ "nostdkey"  , 'k', POPT_ARGFLAG_OR, &arg_nokeywords , 0,
	  "dont include stdkeywords", "level" },
	{ "savelongcomment"  , 'c', POPT_ARGFLAG_OR, &arg_savelongcomment , 0,
	  "dont remove /** comments", "" },
	{ "linerefs"  , 0, POPT_ARGFLAG_OR, &arg_linerefs , 0,
	  "include #line statements", "" },
	{ "suniq"  , 0, POPT_ARGFLAG_OR, &arg_simpleuniq , 0,
	  "user simple identifiers for substitution", "" },
	{ "verbose"  , 'v', POPT_ARGFLAG_OR, &arg_verbose , 0,
	  "be more verbose", "level" },
/*	{ "flag" , 'f', POPT_ARGFLAG_OR, &arg_flag, 0,
	  "flag description", "" },*/
	POPT_AUTOHELP
	{ NULL, 0, 0, NULL, 0, NULL, NULL }
    };
    
    optCon = poptGetContext(NULL, argc,(const char **) argv, optionsTable, 0);

    /* Now do options processing, get portname */
    while ((c = poptGetNextOpt(optCon)) >= 0) {
    }
    if (c < -1) {
	/* an error occurred during option processing */
	fprintf(stderr, "%s: %s\n",
		poptBadOption(optCon, POPT_BADOPTION_NOALIAS),
		poptStrerror(c));
	poptPrintHelp(optCon, stderr, 0);
	exit(1);
    }

    if ((str = poptGetArg(optCon))) {
	arg_infile = (char *)str;
    }
    
    if ((str = poptGetArg(optCon))) {
	fprintf(stderr, "Unknown option '%s'\n", str);
	poptPrintHelp(optCon, stderr, 0);
	exit(1);
    }
    poptFreeContext(optCon);
}

void next_cur(void)
{
    unsigned int ch;
    memmove(&cur[0], &cur[1], sizeof(cur) - 1);
    ch = fgetc(inp);
    if (ch <= 255) {
	cur[sizeof(cur) - 1] = ch;
    } else {
	cur[sizeof(cur) - 1] = 0;
    }
    if (cur[0] == '\n') current_line++;
}

void init_cur(void)
{
    unsigned int i;
    for (i = 0; i < sizeof(cur); i++) {
	next_cur();
    }
}

void echo(void)
{
    if (cur[0] != 0) {
	fputc(cur[0], out);
    }
}

typedef struct subs_s {
    char *from;
    char *to;
} subs_t;

char *unique(char *from)
{
    static unsigned int u = 0;
    static char un[5];
    char *names = "jJyYbBcCdDoOpPqQrRuUiIlL";
    if (!arg_simpleuniq) {
	un[0] = names[u % 24];
	un[1] = names[(u / 24) % 24];
	un[2] = names[(u / (24 * 24)) % 24];
	un[3] = names[(u / (24 * 24 * 24)) % 24];
	un[4] = 0;
	u++;
	return un;
    } else {
	char *ret = malloc(strlen(from) + 7);
	strcpy(ret, "_x_");
	strcat(ret, from);
	strcat(ret, "_x_");
	return ret;
    }
}

subs_t *substitutions = NULL;
unsigned int nsubstitutions = 0;

int sub_compare(const void *_a1, const void *_a2)
{
    const subs_t *a1 = _a1;
    const subs_t *a2 = _a2;

    return strcmp(a1->from, a2->from);
}


char *get_subs(char *from)
{
    subs_t comp;
    subs_t *res;
    comp.from = from;
    
    if (!substitutions) return NULL;

    res = bsearch(&comp, substitutions, nsubstitutions,
		  sizeof(substitutions[0]), sub_compare);
    return res ? res->to : NULL;
}
    

void add_subs(char *from, char *to)
{
    char *tmp;
    subs_t *s;
    subs_t n;
    unsigned int i;

    if ((tmp = get_subs(from))) {
	if (to && !strcmp(tmp, to)) {
	    fprintf(stderr, "Warning: Ignore redefinition from %s to %s\n",
		    from, to);
	    return;
	}
	return;
    }
    if (!to) to = unique(from);
    if (arg_verbose) fprintf(stderr, "Add %s -> %s to substitution list\n", from, to);

    nsubstitutions++;
    substitutions = realloc(substitutions, sizeof(subs_t) * nsubstitutions);
    
    s = substitutions;
    n.from = strdup(from);
    n.to = strdup(to);
    substitutions[nsubstitutions - 1] = n;
    
    for (i = 0; i < nsubstitutions; i++) {
	if (sub_compare(&n, &substitutions[i]) <= 0) {
	    memmove(&substitutions[i + 1], &substitutions[i],
		    sizeof(subs_t) * (nsubstitutions - i - 1));
	    substitutions[i] = n;
	    break;
	}
    }
}


void parse_comment_line(void)
{
    D fputs("// No comment", out);
    while ((cur[0] != '\n') &&
	   (cur[0] != 0)) {
	next_cur();
    }
}

void parse_comment(void)
{
    D fputs("/* No comment */", out);
    /* skip 2 chars */
    next_cur();next_cur();

    if (arg_savelongcomment && (cur[0] == '*')) {
	fputs("/*", out);
	while (strncmp("*/", cur, 2) &&
	       (cur[0] != 0)) {
	    echo();
	    next_cur();
	}
	next_cur();
	next_cur();
	fputs("*/", out);
    } else {
	while (strncmp("*/", cur, 2) &&
	       (cur[0] != 0)) {
	    next_cur();
	}
	next_cur();
	next_cur();
    }
}

void parse_char(void)
{
    D fputs("/*char <*/", out);
    echo();
    next_cur();
    
    while ((cur[0] != '\'') &&
	   (cur[0] != 0)) {
	if (cur[0] == '\\') {
	    echo();
	    next_cur();
	}
	echo();
	next_cur();
    }
    echo();
    next_cur();
    D fputs("/*> char*/", out);
}

void parse_string(void)
{
    D fputs("/*str <*/", out);
    echo();
    next_cur();
    
    while ((cur[0] != '"') &&
	   (cur[0] != 0)) {
	if (cur[0] == '\\') {
	    echo();
	    next_cur();
	}
	echo();
	next_cur();
    }
    echo();
    next_cur();
    D fputs("/*> str*/", out);
}

void parse_ws(void)
{
    fputs(" ", out);
    while (((cur[0] == ' ') ||
	    (cur[0] == '\t')) &&
	   (cur[0] != 0)) {
	next_cur();
    }
}

void parse_ws_nl(void)
{
    int newlines = 0;
    static int first = 1;
    fputs("\n", out);
    while (((cur[0] == ' ') ||
	    (cur[0] == '\t') ||
	    (cur[0] == '\n')) &&
	   (cur[0] != 0)) {
	if (cur[0] == '\n') newlines++;
	next_cur();
    }

    if (arg_linerefs && ((newlines > 1) || first))
	fprintf(out, "#line %d \"%s\"\n", current_line, arg_infile ? arg_infile : "(stdin)");
    first = 0;
}

void parse_ident()
{
    char buf[MAX_IDENT_LEN];
    unsigned int pos;

    buf[MAX_IDENT_LEN - 1] = 0;
    for (pos = 0; pos < MAX_IDENT_LEN - 1; pos++) {
	buf[pos] = cur[0];
	next_cur();
	switch (cur[0]) {
	case '0' ... '9':
	case 'a' ... 'z':
	case 'A' ... 'Z':
	case '_':
	    continue;
	}
	buf[pos + 1] = 0;
	break;
	
    }

    //D fputs("/*ident<*/", out);
    {
	char *subst = get_subs(buf);

	fputs(subst ? subst : buf, out);

	if (left_over_file && !subst) {
	    fputs(buf, left_over_file);
	    fputs("\t", left_over_file);
	    fputs(buf, left_over_file);
	    fputs("\n", left_over_file);
	    add_subs(buf, buf);
	}
    }
    //D fputs("/*>ident*/", out);
    

}

void parse(void)
{
    init_cur();
    while (1) {
	switch (cur[0]) {
	case '/':
	    if (!strncmp("//", cur, 2)) {
		parse_comment_line();
	    } else if (!strncmp("/*", cur, 2)) {
		parse_comment();
	    } else {
		echo();
		next_cur();
	    };
	    break;
	case '\'':
	    parse_char();
	    break;
	case '"':
	    parse_string();
	    break;
	case 'a' ... 'z':
	case 'A' ... 'Z':
	case '_':
	    parse_ident();
	    break;
	case ' ':
	case '\t':
	    parse_ws();
	    break;
	case '\n':
	    parse_ws_nl();
	    break;
	case '\\':
	    if (cur[1] == '\n') {
		/* quoted newline */
		echo(); next_cur();
	    }
	    echo(); next_cur();
	    break;
	case 0:
	    goto out; /* eof */
	default:
	    echo();
	    next_cur();
	    break;
	}
    }
 out:
    ;
}

FILE *fopen_checked(char *filename, char *mode)
{
    FILE *ret = fopen(filename, mode);
    if (!ret) {
	fprintf(stderr, "Cant open file %s for %s : ",
		filename, strcmp(mode, "r") ? "writeing" :"reading");
	perror("");
	exit(1);
    }
    return ret;
}


void read_subs(char *filename)
{
    FILE *s;
    char f[MAX_IDENT_LEN];
    char t[MAX_IDENT_LEN];
    char line[MAX_IDENT_LEN * 2 + 10];
    
    if (!filename) return;

    s = fopen_checked(filename, "r");

    while (!feof(s)) {
	int ret;
	if (fgets(line, sizeof(line), s)) {
	    ret = sscanf(line, "%s %s", f, t);
	    if (ret == 1) {
		add_subs(f, NULL);
	    } else if (ret == 2) {
		add_subs(f, t);
	    } else if (ret == 1) {
		fprintf(stderr, "Scan error in file %s near %s\n",
			filename, f);
		exit(1);
	    } else if (ret == EOF) {
		break;
	    } else {
		fprintf(stderr, "Scan error in file %s (%d)\n",
			filename, ret);
		exit(1);
	    }
	}
	if (ferror(s)) {
	    fprintf(stderr, "Error reading file %s", filename);
	    perror("");
	    exit(1);
	}
    }
    fclose(s);
}

char *reserved_words[] = {
    "_Bool",	
    "_Complex",	
    "__FUNCTION__",
    "__PRETTY_FUNCTION__", 
    "__alignof",	
    "__alignof__",	
    "__asm",	
    "__asm__",	
    "__attribute",	
    "__attribute__",
    "__bounded",	
    "__bounded__",	
    "__builtin_choose_expr",
    "__builtin_types_compatible_p",
    "__builtin_va_arg",	
    "__complex",	
    "__complex__",	
    "__const",	
    "__const__",	
    "__extension__",
    "__func__",	
    "__imag",	
    "__imag__",	
    "__inline",	
    "__inline__",	
    "__label__",	
    "__ptrbase",	
    "__ptrbase__",	
    "__ptrextent",	
    "__ptrextent__",
    "__ptrvalue",	
    "__ptrvalue__",
    "__real",	
    "__real__",	
    "__restrict",	
    "__restrict__",
    "__signed",	
    "__signed__",	
    "__thread",	
    "__typeof",	
    "__typeof__",	
    "__unbounded",	
    "__unbounded__",
    "__volatile",	
    "__volatile__",
    "asm",		
    "auto",	
    "break",	
    "case",	
    "char",	
    "const",	
    "continue",	
    "default",	
    "do",		
    "double",	
    "else",	
    "enum",	
    "extern",	
    "float",	
    "for",		
    "goto",	
    "if",		
    "inline",	
    "int",		
    "long",	
    "register",	
    "restrict",	
    "return",	
    "short",	
    "signed",	
    "sizeof",	
    "static",	
    "struct",	
    "switch",	
    "typedef",	
    "typeof",	
    "union",	
    "unsigned",	
    "void",	
    "volatile",
    "while",

    "include",
    "define",
    "defined",
    "ifdef",
    "ifndef",
    "elif",
    NULL
};

void read_keywords(void)
{
    char **str;

    for (str = reserved_words; *str; str++) {
	add_subs(*str, *str);
    }
    
}

/* main */
int main(int argc, char **argv)
{
    inp = stdin;
    out = stdout;
    parse_opt(argc, argv);

    if (!arg_nokeywords) read_keywords();
    
    if (arg_subs_left_file) {
	left_over_file = fopen_checked(arg_subs_left_file, "w");
    }

    inp = arg_infile ? fopen_checked(arg_infile, "r") : stdin;
    out = arg_outfile ? fopen_checked(arg_outfile, "w") : stdout;
    
    read_subs(arg_subs_file);

    parse();
    if (left_over_file) fclose(left_over_file);
    return 0;
}


/* Example -s file for myself :
stdio	stdio
h	h
stdlib	stdlib
unistd	unistd
errno	errno
string	string
signal	signal
math	math
popt	popt
FILE	FILE
inp
out
left_over_file
NULL	NULL
arg_verbose
arg_nokeywords
arg_subs_file
arg_subs_left_file
D
MAX_IDENT_LEN
cur
next_cur
ch
memmove	memmove
fgetc fgetc
init_cur
i
echo
fputc	fputc
subs_s
from
to
subs_t
unique
u
un
names
substitutions
nsubstitutions
sub_compare
_a1
_a2
a1
a2
strcmp	strcmp
get_subs
comp
res
bsearch	bsearch
add_subs
tmp
s
n
fprintf	fprintf
stderr	stderr
realloc	realloc
strdup	strdup
parse_comment_line
fputs	fputs
parse_comment
strncmp	strncmp
parse_char
parse_string
parse_ident
buf
pos
subst
parse
parse_opt
argc
argv
str
c
POPT_ARGFLAG_OR	POPT_ARGFLAG_OR
POPT_AUTOHELP	POPT_AUTOHELP
poptGetContext	poptGetContext
poptGetNextOpt	poptGetNextOpt
poptBadOption	poptBadOption
POPT_BADOPTION_NOALIAS	POPT_BADOPTION_NOALIAS
poptStrerror	poptStrerror
poptPrintHelp	poptPrintHelp
exit	exit
poptGetArg	poptGetArg
poptFreeContext	poptFreeContext
read_subs
filename
f
t
line
fopen	fopen
perror	perror
feof	feof
ret
fgets	fgets
sscanf	sscanf
EOF	EOF
ferror	ferror
fclose	fclose
reserved_words
read_keywords
main	main
stdin	stdin
stdout	stdout
*/ 


/*
 * Local Variables:
 *  compile-command: "gcc cscrambler.c -Wall -W -Wno-unused -lpopt -g -O2 -o cscrambler"
 * End:
 *
 */
