/*
 *               ParaStation
 *
 * Copyright (C) 2004 ParTec AG, Karlsruhe
 * Copyright (C) 2005 Cluster Competence Center GmbH, Munich
 *
 * $Id: pscom.c 5307 2007-10-24 14:46:02Z hauke $
 *
 */
/**
 * @author
 *         Jens Hauke <hauke@par-tec.de>
 *
 * @file
 ***********************************************************/

#include "pscom.h"
#include "pscom_priv.h"

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <ctype.h>
#include <assert.h>

#include <fcntl.h>
#include <sched.h>
#include <netdb.h>
#include <netinet/in.h>

#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/poll.h>

#include "list.h"
#include "getid.c"
#include "pscom_ufd.h"
#include "pscom_str_util.h"
#include "pscom_con.h"
#include "pscom_env.h"

pscom_t pscom = {
	/* parameter from environment */
	.env = {
		.debug = 0,
		.so_sndbuf = 32768,
		.so_rcvbuf = 32768,
		.tcp_nodelay = 1,

		.sched_yield = 0,
		.rendezvous_size = ~0,
		.rendezvous_size_shm = ~0, /* default rendezvous_size for shm */
		.rendezvous_size_dapl = ~0, /* default rendezvous_size for dapl */
		.rendezvous_size_elan = ~0, /* default rendezvous_size for elan */
		.sigquit = 0,
		.readahead = 100,
		.skipblocksize = 8192,
		.retry = 4,

		.plugindir = "",
	},
	/* statistic */
	.stat = {
		.reqs = 0,
		.gen_reqs = 0,
		.gen_reqs_used = 0,
		.progresscounter = 0,
		.progresscounter_check = 0,
	},
};


int pscom_writeall(int fd, const void *buf, int count)
{
	int len;
	int c = count;
	
	while (c > 0) {
		len = write(fd, buf, c);
		if (len < 0) {
			if ((errno == EINTR) || (errno == EAGAIN)) {
				sched_yield();
				continue;
			} else
				return -1;
		}
		c -= len;
		buf = ((char*)buf) + len;
	}
	
	return count;
}


int pscom_readall(int fd, void *buf, int count)
{
	int len;
	int c = count;
	
	while (c > 0) {
		len = read(fd, buf, c);
		if (len <= 0) {
			if (len < 0) {
				if ((errno == EINTR) || (errno == EAGAIN)) {
					sched_yield();
					continue;
				} else
					return -1;
			} else {
				return count - c;
			}
		}
		c -= len;
		buf = ((char*)buf) + len;
	}
	
	return count;
}


#define PSP_NDCBS 10
void pscom_unlock(void)
{
	int ndcbs;
	int i;
	int more;
	struct {
		pscom_req_t *req;
	} dcbs[PSP_NDCBS];
	
restart:
	more = 0;
	ndcbs = 0;
	if (list_empty(&pscom.io_doneq)) {
		_pscom_unlock();
		return;
	}

	
	while (!list_empty(&pscom.io_doneq)) {
		pscom_req_t *req = list_entry(pscom.io_doneq.next, pscom_req_t, next);
		list_del(&req->next);
		dcbs[ndcbs].req = req;
		ndcbs++;
		if (ndcbs == PSP_NDCBS) {
			more = 1;
			break;
		}
	}

	_pscom_unlock();
	
	/* execute the done callbacks (without any lock) */
	for(i = 0; i < ndcbs; i++) {
		pscom_req_t *req = dcbs[i].req;

		req->pub.state |= PSCOM_REQ_STATE_DONE;
		req->pub.ops.io_done(&req->pub);
		// do not use req after io_done()! io_done could call free(req)
	}
	if (more) {
		/* There are more requests left */
		pscom_lock();
		goto restart;
	}
}


void pscom_poll_write_stop(pscom_con_t *con)
{
	/* it's save to dequeue more then once */
	list_del_init(&con->poll_next_send);
}


void pscom_poll_write_start(pscom_con_t *con)
{
	if (list_empty(&con->poll_next_send)) {
		list_add_tail(&con->poll_next_send, &pscom.poll_sender);
	}
	con->do_write(con);
	/* Dont do anything after this line.
	   do_write() can reenter pscom_poll_write_start()! */
}


void pscom_poll_read_start(pscom_con_t *con)
{
	pscom_poll_reader_t *reader = &con->poll_reader;
	if (list_empty(&reader->next)) {
		list_add_tail(&reader->next, &pscom.poll_reader);
	}

	reader->do_read(reader);
	/* Dont do anything after this line.
	   do_read() can reenter pscom_poll_read_start()! */
}


void pscom_poll_read_stop(pscom_con_t *con)
{
	pscom_poll_reader_t *reader = &con->poll_reader;

	/* it's save to dequeue more then once */
	list_del_init(&reader->next);
}


int pscom_progress(int timeout)
{
	struct list_head *pos, *next;

	list_for_each_safe(pos, next, &pscom.poll_sender) {
		pscom_con_t *con = list_entry(pos, pscom_con_t, poll_next_send);
		con->do_write(con);
		timeout = 0; // enable polling
	}

	list_for_each_safe(pos, next, &pscom.poll_reader) {
		pscom_poll_reader_t *reader = list_entry(pos, pscom_poll_reader_t, next);
		if (reader->do_read(reader)) {
			return 1;
		}
		timeout = 0; // enable polling
	}

	if (ufd_poll(&pscom.ufd, timeout)) {
		return 1;
	} else {
		if (pscom.env.sched_yield) {
			sched_yield();
		}
		return 0;
	}
}


static
void pscom_cleanup(void)
{
	DPRINT(3,"pscom_cleanup()");
	while (!list_empty(&pscom.sockets)) {
		pscom_sock_t *sock = list_entry(pscom.sockets.next, pscom_sock_t, next);
		pscom_close_socket(&sock->pub);
	}
	pscom_plugins_destroy();
	DPRINT(1,"Byee.");
}

/*
******************************************************************************
*/

void pscom_set_debug(unsigned int level)
{
	pscom.env.debug = level;
}


int pscom_init(int pscom_version)
{
	static int init=0;

	if (((pscom_version & 0xff00) != (PSCOM_VERSION & 0xff00)) ||
	    (pscom_version > PSCOM_VERSION)) {
		// different major number, or minor number bigger
		// (new libs support old api, if major number is equal)
		return PSCOM_ERR_UNSUPPORTED_VERSION;
	}
	if (init)
		return PSCOM_SUCCESS;
	init = 1;

	{
		int res_mutex_init;
		res_mutex_init = pthread_mutex_init(&pscom.global_lock, NULL); 
		assert(res_mutex_init == 0);
		res_mutex_init = pthread_mutex_init(&pscom.lock_requests, NULL); 
		assert(res_mutex_init == 0);
	}

	ufd_init(&pscom.ufd);
	INIT_LIST_HEAD(&pscom.sockets);
	INIT_LIST_HEAD(&pscom.requests);
	INIT_LIST_HEAD(&pscom.io_doneq);

	INIT_LIST_HEAD(&pscom.poll_reader);
	INIT_LIST_HEAD(&pscom.poll_sender);
	
	pscom_env_init();

	pscom_debug_init();

	atexit(pscom_cleanup);
	return PSCOM_SUCCESS;
}


int pscom_get_nodeid(void)
{
	static uint32_t id = 0;
	/*  p4s_node_id(void) expect the IP of this node! */

	if (!id) {
		id = psp_getid(); /* Use env PSP_NETWORK to get an IP */
	}
	return id;
}


int pscom_get_portno(pscom_socket_t *socket)
{
	return socket->listen_portno;
}


int pscom_test_any(void)
{
	int ret;
	pscom_lock(); {
		ret = pscom_progress(0);
	} pscom_unlock();

	return ret;
}


void pscom_wait_any(void)
{
	pscom_lock(); {
		if (pscom.stat.progresscounter ==
		    pscom.stat.progresscounter_check) {
			pscom_progress(-1);
		} else {
			pscom_progress(0);
			pscom.stat.progresscounter_check = pscom.stat.progresscounter;
		}
	} pscom_unlock();
}
