/**************************************************** -*- linux-c -*- *
 *                  ParaStation
 *
 * Copyright (c) 2007 ParTec Cluster Competence Center GmbH
 * All rights reserved.
 **********************************************************************/
/**
 * $Id: pscom_con.c 5310 2007-10-24 15:26:35Z hauke $
 *
 * @author
 *         Jens Hauke <hauke@par-tec.de>
 *
 **********************************************************************/
#include "pscom_con.h"
#include "pscom_str_util.h"
#include "pscom_io.h"
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <netinet/tcp.h>
#include <errno.h>

static
int mtry_connect(int sockfd, const struct sockaddr *serv_addr,
		 socklen_t addrlen)
{
/* In the case the backlog (listen) is smaller than the number of
   processes, the connect could fail with ECONNREFUSED even though
   there is a linstening socket. mtry_connect() retry four times
   the connect after one second delay.
*/
	unsigned int i;
	int ret = 0;
	struct sockaddr_in *sa = (struct sockaddr_in*)serv_addr;
	for (i = 0; i < pscom.env.retry; i++) {
		ret = connect(sockfd, serv_addr, addrlen);
		if (ret >= 0) break;
		if (errno != ECONNREFUSED) break;
		sleep(1);
		DPRINT(2, "Retry %d CONNECT to %s:%d",
		       i + 1,
		       pscom_inetstr(ntohl(sa->sin_addr.s_addr)),
		       ntohs(sa->sin_port));
	}
	return ret;
}


static
void pscom_no_rw_start_stop(pscom_con_t *con)
{
}


static
void tcp_configure(int fd)
{
	int ret;
	int val;
	
	if (pscom.env.so_sndbuf) {
		errno = 0;
		val = pscom.env.so_sndbuf;
		ret = setsockopt(fd, SOL_SOCKET, SO_SNDBUF, &val, sizeof(val));
		DPRINT(2, "setsockopt(%d, SOL_SOCKET, SO_SNDBUF, [%d], %ld) = %d : %s",
		       fd, val, (long)sizeof(val), ret, strerror(errno));
	}
	if (pscom.env.so_rcvbuf) {
		errno = 0;
		val = pscom.env.so_rcvbuf;
		ret = setsockopt(fd, SOL_SOCKET, SO_RCVBUF, &val, sizeof(val));
		DPRINT(2, "setsockopt(%d, SOL_SOCKET, SO_RCVBUF, [%d], %ld) = %d : %s",
		       fd, val, (long)sizeof(val), ret, strerror(errno));
	}
	errno = 0;
	val = pscom.env.tcp_nodelay;
	ret = setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, &val, sizeof(val));
	DPRINT(2, "setsockopt(%d, IPPROTO_TCP, TCP_NODELAY, [%d], %ld) = %d : %s",
	       fd, val, (long) sizeof(val), ret, strerror(errno));
	
	ret = fcntl(fd, F_SETFL, O_NONBLOCK);
	if (ret == -1) {
		DPRINT(1, "tcp_configure(): fcntl(%d, F_SETFL, O_NONBLOCK) failed : %s",
		       fd, strerror(errno));
	}
}


// clear sendq. finish all send requests with error
static
void _pscom_con_terminate_sendq(pscom_con_t *con)
{
	while (!list_empty(&con->sendq)) {
		pscom_req_t *req = list_entry(con->sendq.next, pscom_req_t, next);

		list_del(&req->next); // dequeue
		
		req->pub.state |= PSCOM_REQ_STATE_ERROR;
		_pscom_send_req_done(req); // done
	}
}


// clear all recvq's of this connection. finish all recv requests
// of this connection with error. (keep recv any!)
void pscom_con_terminate_recvq(pscom_con_t *con)
{
	struct list_head *pos, *next;

	// current receive:
	if (con->in.req) {
		pscom_req_t *req = con->in.req;
		con->in.req = NULL;
		req->pub.state |= PSCOM_REQ_STATE_ERROR;
		_pscom_recv_req_done(req); // done
	}
	
	// Recv Queue:
	while (!list_empty(&con->recvq)) {
		pscom_req_t *req = list_entry(con->recvq.next, pscom_req_t, next);

		list_del(&req->next); // dequeue
		
		req->pub.state |= PSCOM_REQ_STATE_ERROR;
		_pscom_recv_req_done(req); // done
	}

	// RecvAny Queue:
	list_for_each_safe(pos, next, &get_sock(con->pub.socket)->recvq_any) {
		pscom_req_t *req = list_entry(pos, pscom_req_t, next);

//		fprintf(stderr, "Test rm "RED"req %p  con %p == %p "NORM"\n", req, req->pub.connection, &con->pub);
		if (req->pub.connection == &con->pub) {
//			fprintf(stderr, RED "remove con %p\n"NORM, req);

			list_del(&req->next);
			req->pub.state |= PSCOM_REQ_STATE_ERROR;
			_pscom_recv_req_done(req); // done
		}
	}
}


static
void _pscom_con_terminate_genreqq(pscom_con_t *con)
{
	// genreqq:
	while (!list_empty(&con->genrecvq)) {
		pscom_req_t *req = list_entry(con->genrecvq.next, pscom_req_t, next);
		
		_genreq_dequeue(req);

		/* ToDo: if (genreq->partner_req) send rendezvous Cancel!!! */
		req->partner_req = NULL;
		
		_pscom_grecv_req_done(req); // done
	}
}


static
void pscom_con_end_read(pscom_con_t *con)
{
	con->read_stop(con);
	con->pub.state &= ~PSCOM_CON_STATE_R; // clear R
	con->read_start = pscom_con_terminate_recvq; 

	pscom_con_terminate_recvq(con);
}


static
void pscom_con_end_write(pscom_con_t *con)
{
	con->write_stop(con);
	con->pub.state &= ~PSCOM_CON_STATE_W; // clear W
	con->write_start = _pscom_con_terminate_sendq;
		
	_pscom_con_terminate_sendq(con);
}


static
void pscom_con_error_read_failed(pscom_con_t *con, pscom_err_t error)
{
	pscom_con_end_read(con);
}


static
void pscom_con_error_write_failed(pscom_con_t *con, pscom_err_t error)
{
	pscom_con_end_write(con);
}


void pscom_con_close(pscom_con_t *con)
{
	assert(con->magic == MAGIC_CONNECTION);
	if (con->pub.state != PSCOM_CON_STATE_CLOSED) {
		D_TR(printf("pscom_con_close(con:%p) : state: %s\n", con,
			    pscom_con_state_str(con->pub.state)));
	retry:
		pscom_con_end_write(con);
		pscom_con_end_read(con);
		
		_pscom_con_terminate_genreqq(con);
		
		assert(con->pub.state == PSCOM_CON_STATE_NO_RW);
		assert(list_empty(&con->sendq));
		assert(list_empty(&con->recvq));
		assert(list_empty(&con->genrecvq));
		assert(con->in.req == NULL);

		pscom_call_io_done();

		if (!list_empty(&con->sendq) ||
		    !list_empty(&con->recvq) ||
		    !list_empty(&con->genrecvq) ||
		    con->in.req) goto retry; // in the case the io_donq callbacks post more work

		if (con->close) con->close(con);
		
		list_del(&con->next);
		con->pub.state = PSCOM_CON_STATE_CLOSED;

		_pscom_step();
	}
}


void pscom_con_error(pscom_con_t *con, pscom_op_t operation, pscom_err_t error)
{
	assert(con->magic == MAGIC_CONNECTION);

	DPRINT(error != PSCOM_ERR_EOF ? 1 : 2,
	       "connection to %s (type:%s,state:%s) : %s : %s",
	       pscom_con_info_str(&con->pub.remote_con_info),
	       pscom_con_type_str(con->pub.type),
	       pscom_con_state_str(con->pub.state),
	       pscom_op_str(operation),
	       pscom_err_str(error));

	_pscom_step();

	switch (operation) {
	case PSCOM_OP_READ: 
		pscom_con_error_read_failed(con, error);
		break;
	case PSCOM_OP_WRITE:
		pscom_con_error_write_failed(con, error);
		break;
	}

	if (con->pub.socket->ops.con_error) {
		con->pub.socket->ops.con_error(&con->pub, operation, error);
	}
}


static
void pscom_con_info(pscom_con_t *con, pscom_con_info_t *con_info)
{
	*con_info = con->pub.socket->local_con_info;
	con_info->id = &con->pub;
}


static
pscom_con_t *pscom_con_create(pscom_sock_t *sock)
{
	pscom_con_t *con;
	con = malloc(sizeof(*con) + sock->pub.connection_userdata_size);
	if (!con) return NULL;
	
	con->magic = MAGIC_CONNECTION;
	con->pub.socket = &sock->pub;
	con->pub.userdata_size = sock->pub.connection_userdata_size;
	con->pub.state = PSCOM_CON_STATE_CLOSED;
	con->pub.type = PSCOM_CON_TYPE_NONE;
	
	INIT_LIST_HEAD(&con->sendq);
	INIT_LIST_HEAD(&con->recvq);
	INIT_LIST_HEAD(&con->genrecvq);
	INIT_LIST_HEAD(&con->rma_readq);
	
	INIT_LIST_HEAD(&con->poll_reader.next);
	INIT_LIST_HEAD(&con->poll_next_send);
		
	con->in.req	= 0;
	con->in.skip	= 0;

	con->in.readahead.iov_base = NULL;
	con->in.readahead.iov_len = 0;
	con->in.readahead_size = 0;

	con->write_start = pscom_no_rw_start_stop;
	con->write_stop = pscom_no_rw_start_stop;
	con->read_start = pscom_no_rw_start_stop;
	con->read_stop = pscom_no_rw_start_stop;
	con->poll_reader.do_read = NULL;
	con->do_write = NULL;
	con->close = pscom_no_rw_start_stop;
	/* RMA */
	con->rma_mem_register = NULL;
	con->rma_mem_deregister = NULL;
	con->rma_read = NULL;

	con->rendezvous_size = pscom.env.rendezvous_size;

	return con;
}


static
void pscom_con_destroy(pscom_con_t *con)
{
	assert(con->magic == MAGIC_CONNECTION);
	assert(con->pub.state == PSCOM_CON_STATE_CLOSED);
	assert(list_empty(&con->poll_next_send));
	assert(list_empty(&con->poll_reader.next));
	
	con->magic = 0;
	free(con);
}


static
void pscom_con_setup(pscom_con_t *con)
{
	/* ToDo: disable read, if no "any" and no "node specific" request
	   is enqueued */
	con->read_start(con);
    
	if (!list_empty(&con->sendq)) {
		con->write_start(con);
	}
}


static
int pscom_is_valid_con(pscom_con_t *con)
{
	struct list_head *pos_sock;
	struct list_head *pos_con;
	list_for_each(pos_sock, &pscom.sockets) {
		pscom_sock_t *sock = list_entry(pos_sock, pscom_sock_t, next);
		
		list_for_each(pos_con, &sock->connections) {
			pscom_con_t *con2 = list_entry(pos_con, pscom_con_t, next);

			if (con2 == con) {
				D_TR(printf("pscom_is_valid_con(%p) = 1\n", con));
				return 1;
			}
		}
	}
	D_TR(printf("pscom_is_valid_con(%p) = 0\n", con));

	return 0;
}

	

void pscom_con_accept(ufd_t *ufd, ufd_funcinfo_t *ufd_info)
{
	pscom_sock_t *sock = ufd_info->priv;

	pscom_con_t *con;
	int con_fd;
	pscom_con_info_t con_info;

	/* Open connection */
	con = pscom_con_create(sock);
	if (!con) goto err_connection_create;

	/* Open the socket */
	con_fd = accept(sock->listen_fd, NULL, NULL);
	if (con_fd < 0) goto err_accept;

	tcp_configure(con_fd);

	/* exchange connection information */
	pscom_con_info(con, &con_info);
	
	pscom_readall(con_fd, &con->pub.remote_con_info, sizeof(pscom_con_info_t));
	pscom_writeall(con_fd, &con_info, sizeof(pscom_con_info_t));

	while (1) {
		int arch;

		if (pscom_readall(con_fd, &arch, sizeof(arch)) != sizeof(arch))
			goto err_init_failed;

		pscom_plugin_t *p = pscom_plugin_by_archid(arch);

		if (p) {
			if (p->con_accept(con, con_fd)) goto out;
		} else {
			// Unknown or disabled arch
			arch = PSCOM_ARCH_ERROR;
			pscom_writeall(con_fd, &arch, sizeof(arch));
		}
	}
	/* --- */
out:
	{
		char *fs = strdup(pscom_con_info_str(&con_info));
		DPRINT(1, "ACCEPT %s to %s via %s",
		       pscom_con_info_str(&con->pub.remote_con_info), fs,
		       pscom_con_type_str(con->pub.type));
		free(fs);
	}

	list_add_tail(&con->next, &sock->connections);
	
	if (sock->pub.ops.con_accept) {
		pscom_unlock(); {
			sock->pub.ops.con_accept(&con->pub);
		} pscom_lock();
	}

	// warning: sock->pub.ops.connection_accept() can call free(con)!
	if (pscom_is_valid_con(con)) {
		pscom_con_setup(con);
	}

	_pscom_step();

	return;
	/* --- */
err_init_failed:
	errno = EPIPE;
	close(con_fd);
err_accept:
	pscom_con_destroy(con);
	DPRINT(1, "ACCEPT failed : %s", strerror(errno));
	return;
	/* --- */
err_connection_create:
	DPRINT(1, "ACCEPT failed (create connection failed) : %s",
	       strerror(errno));
	return;
}


static
pscom_err_t pscom_con_connect_via_tcp(pscom_con_t *con, struct sockaddr_in *si)
{
	int con_fd;
	int initialized = 0;
	pscom_con_info_t con_info;
	
	/* Open the socket */
	con_fd = socket(PF_INET , SOCK_STREAM, 0);
	if (con_fd < 0) goto err_socket;

	/* Connect */
	if (mtry_connect(con_fd, (struct sockaddr*)si, sizeof(*si)) < 0) goto err_connect;

	tcp_configure(con_fd);

	/* exchange connection information */
	pscom_con_info(con, &con_info);

	pscom_writeall(con_fd, &con_info, sizeof(pscom_con_info_t));
	pscom_readall(con_fd, &con->pub.remote_con_info, sizeof(pscom_con_info_t));

	struct list_head *pos;

	/* Search for "best" connections */
	list_for_each(pos, &pscom_plugins) {
		pscom_plugin_t *p = list_entry(pos, pscom_plugin_t, next);
		initialized = p->con_connect(con, con_fd);
		if (initialized) break;
	}
    
	if (!initialized)
		goto err_init_failed;

	{
		char *fs = strdup(pscom_con_info_str(&con_info));
		DPRINT(1, "CONNECT %s to %s via %s",
		       fs, pscom_con_info_str(&con->pub.remote_con_info),
		       pscom_con_type_str(con->pub.type));
		free(fs);
	}
	
	list_add_tail(&con->next, &get_sock(con->pub.socket)->connections);

	pscom_con_setup(con);

	return PSCOM_SUCCESS;

	/* error code */
err_init_failed:
	errno = EUNATCH;
err_connect:
	close(con_fd);
err_socket:
	DPRINT(1, "CONNECT %s to tcp:%s:%u FAILED : %s",
	       pscom_con_info_str(&con_info),
	       pscom_inetstr(ntohl(si->sin_addr.s_addr)),
	       ntohs(si->sin_port),
	       strerror(errno));

	return PSCOM_ERR_STDERROR;
}


static void
loopback_write_start(pscom_con_t *con)
{
	int i;
	unsigned int len;
	struct iovec iov[2];
	pscom_req_t *req;

	while (1) {
		req = pscom_write_get_iov(con, iov);
		if (!req) break;

		len = 0;
		for (i = 0; i < 2; i++) {
			len += iov[i].iov_len;
			if (iov[i].iov_len) {
				pscom_read_done(con, iov[i].iov_base, iov[i].iov_len);
			}
		}
		pscom_write_done(con, req, len);
	}
}


static
pscom_err_t pscom_con_connect_loopback(pscom_con_t *con)
{
	pscom_sock_t *sock = get_sock(con->pub.socket);
	
	/* exchange connection information */
	pscom_con_info(con, &con->pub.remote_con_info);

	con->pub.state = PSCOM_CON_STATE_RW;
	con->pub.type = PSCOM_CON_TYPE_LOOP;
	
	con->write_start = loopback_write_start;
	
	{
		char *fs = strdup(pscom_con_info_str(&con->pub.remote_con_info));
		DPRINT(1, "CONNECT %s to %s via %s",
		       fs, pscom_con_info_str(&con->pub.remote_con_info),
		       pscom_con_type_str(con->pub.type));
		free(fs);
	}
	
	list_add_tail(&con->next, &sock->connections);

	if (sock->pub.ops.con_accept) {
		pscom_unlock(); {
			sock->pub.ops.con_accept(&con->pub);
		} pscom_lock();
	}

	// warning: sock->pub.ops.connection_accept() can call free(con)!
	if (pscom_is_valid_con(con)) {
		pscom_con_setup(con);
	}

	return PSCOM_SUCCESS;
}

/*
******************************************************************************
*/

pscom_connection_t *pscom_open_connection(pscom_socket_t *socket)
{
	pscom_sock_t *sock = get_sock(socket);
	pscom_con_t *con;

	pscom_lock(); {
		con = pscom_con_create(sock);
	} pscom_unlock();

	return con ? &con->pub : NULL;
}


pscom_err_t pscom_connect(pscom_connection_t *connection, int nodeid, int portno)
{
	struct sockaddr_in si;
	pscom_con_t *con = get_con(connection);
	pscom_err_t rc;
	
	/* Initial connection via TCP */
	si.sin_family = PF_INET;
	si.sin_port = htons(portno);
	si.sin_addr.s_addr = htonl(nodeid);
		
	pscom_lock(); {
		if (((nodeid == pscom_get_nodeid()) || (nodeid == -1)) &&
		    ((portno == get_sock(connection->socket)->pub.listen_portno) ||
		     (portno == -1))) {
			rc = pscom_con_connect_loopback(con);
		} else {
			rc = pscom_con_connect_via_tcp(con, &si);
		}
	} pscom_unlock();

	return rc;
}


void pscom_close_connection(pscom_connection_t *connection)
{
	pscom_lock(); {
		pscom_con_t *con = get_con(connection);
		pscom_con_close(con);
		pscom_con_destroy(con);
	} pscom_unlock();
}


pscom_connection_t *pscom_get_next_connection(pscom_socket_t *socket, pscom_connection_t *connection)
{
	pscom_sock_t *sock = get_sock(socket);
	pscom_con_t *res;

	assert(sock->magic == MAGIC_SOCKET);
		
	pscom_lock(); {
		if (!connection) {
			
			if (list_empty(&sock->connections)) {
				res = NULL;
			} else {
				res = list_entry(sock->connections.next, pscom_con_t, next);
			}
		} else {
			pscom_con_t *con = get_con(connection);
			assert(con->magic == MAGIC_CONNECTION);
			
			if (con->next.next != &sock->connections) {
				res = list_entry(con->next.next, pscom_con_t, next);
			} else {
				res = NULL;
			}
		}
	} pscom_unlock();

	return res ? &res->pub : NULL;
}
