/**************************************************** -*- linux-c -*- *
 *                  ParaStation
 *
 * Copyright (c) 2007 ParTec Cluster Competence Center GmbH
 * All rights reserved.
 **********************************************************************/
/**
 * $Id: pscom_debug.c 5310 2007-10-24 15:26:35Z hauke $
 *
 * @author
 *         Jens Hauke <hauke@par-tec.de>
 *
 **********************************************************************/

#include "pscom_priv.h"
#include "pscom_debug.h"
#include <signal.h>

#if 1
/* MPI2 Header: */

/* from mpid/psp/include/mpidpre.h */
typedef struct MPID_PSCOM_XHeader {
	int32_t		tag;
	uint16_t	context_id;
	uint8_t		type; /* one of MPID_PSP_MSGTYPE */
	uint8_t		_reserved_;
	int32_t		src_rank;
} MPID_PSCOM_XHeader_t;

/* from mpid/psp/include/mpidpre.h */
enum MPID_PSP_MSGTYPE {
	MPID_PSP_MSGTYPE_DATA,			/* Data message */
	MPID_PSP_MSGTYPE_DATA_REQUEST_ACK,	/* Data message and request DATA_ACK acknowledge */
	MPID_PSP_MSGTYPE_DATA_ACK,		/* Acknowledge of DATA_REQUEST_ACK */
	MPID_PSP_MSGTYPE_CANCEL_DATA_ACK,	/* Acknowledge of CANCEL_DATA_REQUEST_ACK */
	MPID_PSP_MSGTYPE_CANCEL_DATA_REQUEST_ACK, /* Cancel an already send DATA message. Request CANCEL_DATA_ACK. */

	/* One Sided communication: */
	MPID_PSP_MSGTYPE_RMA_PUT,
	MPID_PSP_MSGTYPE_RMA_GET_REQ,
	MPID_PSP_MSGTYPE_RMA_GET_ANSWER,
	MPID_PSP_MSGTYPE_RMA_ACCUMULATE,

	MPID_PSP_MSGTYPE_RMA_LOCK_SHARED_REQUEST,
	MPID_PSP_MSGTYPE_RMA_LOCK_EXCLUSIVE_REQUEST,
	MPID_PSP_MSGTYPE_RMA_LOCK_ANSWER,

	MPID_PSP_MSGTYPE_RMA_UNLOCK_REQUEST,
	MPID_PSP_MSGTYPE_RMA_UNLOCK_ANSWER,
};

/* from mpid/psp/src/mpid_debug.c */
static
const char *mpid_msgtype_str(enum MPID_PSP_MSGTYPE msg_type)
{
	switch (msg_type) {
	case MPID_PSP_MSGTYPE_DATA:		return "DATA";
	case MPID_PSP_MSGTYPE_DATA_REQUEST_ACK:	return "DATA_REQUEST_ACK";
	case MPID_PSP_MSGTYPE_DATA_ACK:		return "DATA_ACK";
	case MPID_PSP_MSGTYPE_CANCEL_DATA_ACK:	return "CANCEL_DATA_ACK";
	case MPID_PSP_MSGTYPE_CANCEL_DATA_REQUEST_ACK: return "CANCEL_DATA_REQUEST_ACK";
	case MPID_PSP_MSGTYPE_RMA_PUT:		return "RMA_PUT";

	case MPID_PSP_MSGTYPE_RMA_GET_REQ:	return "RMA_GET_REQ";
	case MPID_PSP_MSGTYPE_RMA_GET_ANSWER:	return "RMA_GET_ANSWER";
	case MPID_PSP_MSGTYPE_RMA_ACCUMULATE:	return "RMA_ACCUMULATE";
		
	case MPID_PSP_MSGTYPE_RMA_LOCK_SHARED_REQUEST: return "RMA_LOCK_SHARED_REQUEST";
	case MPID_PSP_MSGTYPE_RMA_LOCK_EXCLUSIVE_REQUEST: return "RMA_LOCK_EXCLUSIVE_REQUEST";
	case MPID_PSP_MSGTYPE_RMA_LOCK_ANSWER:	return "RMA_LOCK_ANSWER";
	case MPID_PSP_MSGTYPE_RMA_UNLOCK_REQUEST: return "RMA_UNLOCK_REQUEST";
	case MPID_PSP_MSGTYPE_RMA_UNLOCK_ANSWER: return "RMA_UNLOCK_ANSWER";
	}
	return "UNKNOWN";
}

#endif

#include "perf.c"


static
void pscom_dump_request(FILE *out, pscom_req_t *req)
{
	fprintf(out, "req#%p state:%20s xhlen:%3d dlen:%6d con:%p no:%u\n",
		&req->pub,
		pscom_req_state_str(req->pub.state),
		req->pub.xheader_len,
		req->pub.data_len,
		req->pub.connection,
		req->req_no);
	if (5 <= pscom.env.debug && req->pub.xheader_len >= sizeof(MPID_PSCOM_XHeader_t)) {
		MPID_PSCOM_XHeader_t *xhead = (MPID_PSCOM_XHeader_t *)req->pub.xheader.user;
		fprintf(out, " mpi2: tag:%6d con_id:%4d src_rank:%4d type:%d(%s)\n",
			xhead->tag, xhead->context_id, xhead->src_rank, xhead->type,
			mpid_msgtype_str(xhead->type));
	}
}


static
void pscom_dump_ufd(FILE *out, ufd_t *ufd)
{
	int i;
	for (i = 0; i < ufd->nufds; i++) {
		char buf[20];
		snprintf(buf, sizeof(buf), "%s%d(",
			 i ? "," : "",
			 ufd->ufds[i].fd) ;
		if (ufd->ufds[i].events & POLLIN) strcat(buf, "r");
		if (ufd->ufds[i].events & POLLOUT) strcat(buf, "w");
		if (ufd->ufds[i].events & POLLERR) strcat(buf, "e");
		if (ufd->ufds[i].events & POLLPRI) strcat(buf, "u");
		if (ufd->ufds[i].events & POLLHUP) strcat(buf, "h");
		strcat(buf, ")");
		fputs(buf, out);
	}
	fputs("\n", out);
}


static
void pscom_dump_requests(FILE *out)
{
	struct list_head *pos;

	fprintf(out, "Requests:\n");

	pthread_mutex_lock(&pscom.lock_requests);
	list_for_each(pos, &pscom.requests) {
		pscom_req_t *req = list_entry(pos, pscom_req_t, all_req_next);

		pscom_dump_request(out, req);
	}
	pthread_mutex_unlock(&pscom.lock_requests);
}


static
void pscom_dump_connection(FILE *out, pscom_con_t *con)
{
	fprintf(out, "    con#%p type:%6s state:%8s dest:%s\n", &con->pub,
		pscom_con_type_str(con->pub.type),
		pscom_con_state_str(con->pub.state),
		pscom_con_info_str(&con->pub.remote_con_info));
}


static
void pscom_dump_connections(FILE *out, pscom_sock_t *sock)
{
	struct list_head *pos;
	fprintf(out, "  Connections:\n");
	list_for_each(pos, &sock->connections) {
		pscom_con_t *con = list_entry(pos, pscom_con_t, next);
		pscom_dump_connection(out, con);
	}
}


static
void pscom_dump_socket(FILE *out, pscom_sock_t *sock)
{
	fprintf(out, "  sock#%p listen:%6d                src:%s\n",
		&sock->pub, sock->pub.listen_portno,
		pscom_con_info_str(&sock->pub.local_con_info));
	pscom_dump_connections(out, sock);
}


static
void pscom_dump_sockets(FILE *out)
{
	struct list_head *pos;

	fprintf(out, "Sockets:\n");
	list_for_each(pos, &pscom.sockets) {
		pscom_sock_t *sock = list_entry(pos, pscom_sock_t, next);

		pscom_dump_socket(out, sock);
	}
}


void pscom_dump_reqstat(FILE *out)
{
	fprintf(out, "Reqs:%d GenReqs: (cnt:%d  used:%d)\n",
		pscom.stat.reqs,
		pscom.stat.gen_reqs,
		pscom.stat.gen_reqs_used);
}


void pscom_dump_info(FILE *out)
{
	pscom_dump_requests(out);
	pscom_dump_sockets(out);
	fprintf(out, "Fds:\n");
	pscom_dump_ufd(out, &pscom.ufd);

	pscom_dump_reqstat(out);
}


static
void pscom_sigquit(int sig)
{
	int pid = getpid();
	_DPRINT(" +++++++++ SIGQUIT START ++++ ", pid);
	pscom_dump_info(stderr);
	_DPRINT(" +++++++++ SIGQUIT END ++++++ ", pid);
}


void pscom_debug_init(void)
{
	if (pscom.env.sigquit) {
		signal(SIGQUIT, pscom_sigquit);
	}
}
