/**************************************************** -*- linux-c -*- *
 *                  ParaStation
 *
 * Copyright (c) 2007 ParTec Cluster Competence Center GmbH
 * All rights reserved.
 **********************************************************************/
/**
 * $Id: pscom_io.c 5310 2007-10-24 15:26:35Z hauke $
 *
 * @author
 *         Jens Hauke <hauke@par-tec.de>
 *
 **********************************************************************/

#include "pscom_priv.h"
#include "pscom_io.h"
#include <stdlib.h>
#include <stdio.h>
#include "pscom_str_util.h"
#include "pscom_util.h"

// ToDo: disable:
#define ENABLE_REQUEST_MONITORING 1

#ifdef ENABLE_REQUEST_MONITORING
static
void announce_new_req(pscom_req_t *req)
{
	pthread_mutex_lock(&pscom.lock_requests);
	list_add_tail(&req->all_req_next, &pscom.requests);
	pthread_mutex_unlock(&pscom.lock_requests);
}


static
void announce_free_req(pscom_req_t *req)
{
	pthread_mutex_lock(&pscom.lock_requests);
	list_del(&req->all_req_next);
	pthread_mutex_unlock(&pscom.lock_requests);
}
#else /* ENABLE_REQUEST_MONITORING */
static void announce_new_req(pscom_req_t *req) { }
static void announce_free_req(pscom_req_t *req) { }
#endif

static inline
unsigned int round_up8(unsigned int val)
{
	return (val + 7) & ~7;
}


static
pscom_req_t *pscom_req_create(unsigned int max_xheader_len, unsigned int user_size)
{
	pscom_req_t *req;
	unsigned int max_xhl = pscom_max(round_up8(max_xheader_len), sizeof(req->pub.xheader));
	unsigned int extra_xh_size = max_xhl - sizeof(req->pub.xheader);
	
	req = malloc(sizeof(*req) + extra_xh_size + user_size);
	if (!req) return NULL;
	
	req->magic		= MAGIC_REQUEST;
	req->pub.state		= PSCOM_REQ_STATE_DONE;

	req->pub.xheader_len	= max_xheader_len;
	req->pub.data_len	= 0;
	req->pub.data		= NULL;

	req->pub.connection	= NULL;
	req->pub.socket		= NULL;

	req->pub.ops.recv_accept= NULL;
	req->pub.ops.io_done	= NULL;

	req->pub.user_size	= user_size;
	req->pub.user		= (void*)(((char *)&req->pub.xheader) + max_xhl);

	req->pub.max_xheader_len= max_xhl; // at least sizeof(req->pub.xheader)

	announce_new_req(req);

	req->req_no = ++pscom.stat.reqs; // ToDo: disable debug?
	
	return req;
}


void pscom_req_free(pscom_req_t *req)
{
	assert(req->magic == MAGIC_REQUEST);
	assert(req->pub.state & PSCOM_REQ_STATE_DONE);
	req->magic = 0;
	announce_free_req(req);
	
	free(req);
}


static
void pscom_req_prepare_recv(pscom_req_t *req, const pscom_header_net_t *nh, pscom_connection_t *connection)
{
	unsigned int copy_header = sizeof(req->pub.header) +
		pscom_min(req->pub.xheader_len, nh->xheader_len);

	memcpy(&req->pub.header, nh, copy_header);

	req->cur_data.iov_base = req->pub.data;
	
	if (nh->data_len <= req->pub.data_len) {
		req->cur_data.iov_len = nh->data_len;
		req->skip = 0;
	} else {
		req->cur_data.iov_len = req->pub.data_len;
		req->skip = nh->data_len - req->pub.data_len;
		req->pub.state |= PSCOM_REQ_STATE_TRUNCATED;
	}

	D_TR(printf("%s(req:%p) hlen=%u dlen=%zu dlen_req=%u dlen_net=%u skip=%u\n",
		    __func__, req, copy_header, req->cur_data.iov_len,
		    req->pub.data_len, nh->data_len, req->skip));

	assert(connection);
	req->pub.connection = connection;
}


static inline
unsigned int header_length(pscom_header_net_t *header)
{
	return sizeof(pscom_header_net_t) + header->xheader_len;
}


static inline
int header_complete(void *buf, unsigned int size)
{
	pscom_header_net_t *nhead = (pscom_header_net_t *)buf;

	return (size >= sizeof(pscom_header_net_t)) &&
		(size >= header_length(nhead));
}


static inline
int is_recv_req_done(pscom_req_t *req)
{
	return (req->cur_data.iov_len == 0);
}


static inline
int req_recv_accept(pscom_req_t *req, pscom_connection_t *connection, pscom_header_net_t *header)
{
	int (*recv_accept)(pscom_request_t *request,
			   pscom_connection_t *connection,
			   pscom_header_net_t *header_net);

	D_TR(printf("%s(req:%p)\n", __func__, req));

	recv_accept = req->pub.ops.recv_accept;

	return !recv_accept || recv_accept(&req->pub, connection, header);
}


/*
 * Request queueing user side
 */

static inline
pscom_req_t *_genreq_find_from_con(pscom_con_t *con, pscom_req_t *req)
{
	struct list_head *head = &con->genrecvq;
	struct list_head *pos;
	
	D_TR(printf("%s(con:%p, req:%p)\n", __func__, con, req));

	list_for_each(pos, head) {
		pscom_req_t *genreq = list_entry(pos, pscom_req_t, next);
		
		if (req_recv_accept(req, genreq->pub.connection, &genreq->pub.header)) {
			return genreq;
		}
	}
	return NULL;
}


static inline
pscom_req_t *_genreq_find_from_any(pscom_sock_t *sock, pscom_req_t *req)
{
	struct list_head *pos;
	D_TR(printf("%s(sock:%p, req:%p)\n", __func__, sock, req));

	list_for_each(pos, &sock->genrecvq_any) {
		pscom_req_t *genreq = list_entry(pos, pscom_req_t, gen_next_any);
		
		if (req_recv_accept(req, genreq->pub.connection, &genreq->pub.header)) {
			return genreq;
		}
	}
	return NULL;
}


/* find generated request. */
static inline
pscom_req_t *_genreq_find(pscom_req_t *req)
{
	if (req->pub.connection) {
		return _genreq_find_from_con(get_con(req->pub.connection), req);
	} else {
		// receive "any"
		return _genreq_find_from_any(get_sock(req->pub.socket), req);
	}
}


static inline
unsigned int req_write(pscom_req_t *req, char *buf, unsigned int len)
{
	// printf("req_write() %s\n", pscom_dumpstr(buf, pscom_min(len, 32)));
	if (len <= req->cur_data.iov_len) {
		if (req->cur_data.iov_base != buf) {
			memcpy(req->cur_data.iov_base, buf, len);
		}
		req->cur_data.iov_base += len;
		req->cur_data.iov_len -= len;
		return len;
	} else {
		unsigned int clen = req->cur_data.iov_len;
		unsigned int left;
		if (req->cur_data.iov_base != buf) {
			memcpy(req->cur_data.iov_base, buf, clen);
		}
		req->cur_data.iov_base += clen;
		req->cur_data.iov_len = 0;
		
		left = len - clen;
		
		if (req->skip >= left) {
			req->skip -= left;
			return len;
		} else {
			clen += req->skip;
			req->skip = 0;
			
			return clen;
		}
	}
}


static
void _pscom_rendezvous_read_data(pscom_req_t *user_recv_req, pscom_req_t *rendezvous_req);

	
static
void _genreq_merge(pscom_req_t *newreq, pscom_req_t *genreq)
{
	pscom_con_t *con = get_con(genreq->pub.connection);

//	printf("GHeader: " RED "%s" NORM "\n", pscom_dumpstr(&genreq->pub.header, genreq->pub.xheader_len + sizeof(genreq->pub.header)));
	
	pscom_req_prepare_recv(newreq, &genreq->pub.header, genreq->pub.connection);

	/* copy already received data: */
	
	req_write(newreq, genreq->pub.data, (char*)genreq->cur_data.iov_base - (char *)genreq->pub.data);

	newreq->pub.state |= genreq->pub.state;

	if (con->in.req == genreq) {
		// replace existing genreq by newreq;
		assert((genreq->pub.state & (PSCOM_REQ_STATE_IO_STARTED | PSCOM_REQ_STATE_IO_DONE))
			== PSCOM_REQ_STATE_IO_STARTED);

		con->in.req = newreq;
	} else if (genreq->partner_req) {
		/* genreq from rendezvous. Now request the data: */
		_pscom_rendezvous_read_data(newreq, genreq->partner_req);
		genreq->partner_req = NULL;
	} else {
		/* request done. */
		assert(genreq->pub.state & (PSCOM_REQ_STATE_IO_DONE));
		_pscom_recv_req_done(newreq);
	}
	
	pscom.stat.gen_reqs_used++;

	_pscom_grecv_req_done(genreq);
}


static
void genreq_merge_header(pscom_req_t *newreq, pscom_req_t *genreq)
{
	pscom_req_prepare_recv(newreq, &genreq->pub.header, genreq->pub.connection);
}


static
void _pscom_recv_req_enq_con(pscom_req_t *req, pscom_con_t *con)
{
	list_add_tail(&req->next, &con->recvq);
	con->read_start(con);
}


static
void _pscom_recv_req_enq_any(pscom_req_t *req, pscom_sock_t *sock)
{
	list_add_tail(&req->next, &sock->recvq_any);
	/* ToDo: call con->read_start(con) for every connection! */
}


static
void _pscom_recv_req_enq(pscom_req_t *req)
{
	pscom_sock_t *sock;

	D_TR(printf("%s(req:%p)\n", __func__, req));
	
	if (req->pub.connection) {
		req->pub.socket = req->pub.connection->socket;
	}
	sock = get_sock(req->pub.socket);

	if (list_empty(&sock->recvq_any) && req->pub.connection) {
		_pscom_recv_req_enq_con(req, get_con(req->pub.connection));
	} else {
		_pscom_recv_req_enq_any(req, sock);
	}
}


static inline
void _pscom_recv_req_deq(pscom_req_t *req)
{
	list_del(&req->next);
}


static
void _pscom_rma_read_req_enq(pscom_req_t *req)
{
	pscom_con_t *con = get_con(req->pub.connection);
	list_add_tail(&req->next, &con->rma_readq);
}


static inline
void _pscom_rma_read_req_deq(pscom_req_t *req)
{
	list_del(&req->next);
}


static
void _pscom_send_req_enq(pscom_con_t *con, pscom_req_t *req)
{
//	pscom_con_t *con = get_con(req->pub.connection);

	D_TR(printf("%s(req:%p)\n", __func__, req));

	list_add_tail(&req->next, &con->sendq);
	con->write_start(con);
}


static
void _pscom_send_req_deq(pscom_con_t *con, pscom_req_t *req)
{
	assert(!(req->cur_header.iov_len + req->cur_data.iov_len) || /* io done or */
	       !(req->pub.state & PSCOM_REQ_STATE_IO_STARTED));      /* io not started */
	       
	list_del(&req->next); // dequeue
	if (list_empty(&con->sendq)) {
		con->write_stop(con);
	}
}


static
void pscom_req_prepare_send(pscom_req_t *req, unsigned msg_type)
{
	req->pub.header.msg_type = msg_type;
	req->pub.header.xheader_len = req->pub.xheader_len;
	req->pub.header.data_len = req->pub.data_len;
	
	req->cur_header.iov_base = &req->pub.header;
	req->cur_header.iov_len = sizeof(pscom_header_net_t) + req->pub.header.xheader_len;
	req->cur_data.iov_base = req->pub.data;
	req->cur_data.iov_len = req->pub.data_len;
}


static
void pscom_req_prepare_rma_write(pscom_req_t *req)
{
	req->pub.xheader_len = sizeof(req->pub.xheader.rma_write);
}


/*
 * Request queueing network side
 */

static
void _check_readahead(pscom_con_t *con, size_t len)
{
	if (con->in.readahead_size < len) {
		con->in.readahead.iov_base = realloc(con->in.readahead.iov_base, len);
		con->in.readahead_size = len;
		if (!con->in.readahead.iov_base) {
			perror("allocate mem");
			exit(1);
		}
	}
}


static
void _pscom_update_recvq(pscom_con_t *con)
{
	pscom_req_t *req = con->in.req;
	if (req && is_recv_req_done(req)) {
		con->in.skip = req->skip;
		con->in.req = NULL;
		
		_pscom_recv_req_done(req);
		/* ToDo: Stop receiving  con->set_read(port, con, 0); ?*/
	}
}


static
void _pscom_recvq_any_cleanup(pscom_sock_t *sock)
{
	struct list_head *pos, *next;

	list_for_each_safe(pos, next, &sock->recvq_any) {
		pscom_req_t *req = list_entry(pos, pscom_req_t, next);
		if (req->pub.connection) {
			/* Move request from any queue to con queue */
			_pscom_recv_req_deq(req);
			_pscom_recv_req_enq_con(req, get_con(req->pub.connection));
		} else {
			break;
		}
	}
}


static
pscom_req_t *_pscom_get_posted_recv_req(pscom_con_t *con, pscom_header_net_t *header)
{
	struct list_head *pos;
	pscom_sock_t *sock;
	
	list_for_each(pos, &con->recvq) {
		pscom_req_t *req = list_entry(pos, pscom_req_t, next);

		if (req_recv_accept(req, &con->pub, header)) {
			_pscom_recv_req_deq(req);
			return req;
		}
	}

	sock = get_sock(con->pub.socket);
	
	list_for_each(pos, &sock->recvq_any) {
		pscom_req_t *req = list_entry(pos, pscom_req_t, next);
		if (((!req->pub.connection) || (req->pub.connection == &con->pub)) &&
		    req_recv_accept(req, &con->pub, header)) {
			_pscom_recv_req_deq(req);
			_pscom_recvq_any_cleanup(sock);
			return req;
		}
	}
	return NULL;
}


// for debug:
static
int _is_in_recv_queue(pscom_req_t *req)
{
	struct list_head *pos;
	pscom_con_t *con = get_con(req->pub.connection);
	pscom_sock_t *sock;
	
	list_for_each(pos, &con->recvq) {
		pscom_req_t *qreq = list_entry(pos, pscom_req_t, next);

		if (qreq == req) return 1;
	}

	sock = get_sock(con->pub.socket);
	
	list_for_each(pos, &sock->recvq_any) {
		pscom_req_t *qreq = list_entry(pos, pscom_req_t, next);
		if (qreq == req) return 1;
	}
	
	return 0;
}


static
pscom_req_t *_pscom_generate_recv_req(pscom_con_t *con, pscom_header_net_t *nh)
{
	pscom_req_t *req;

	pscom.stat.gen_reqs++;
    
	req = pscom_req_create(nh->xheader_len, nh->data_len);
	req->pub.state = PSCOM_REQ_STATE_GRECV_REQUEST;
	/* freed inside genreq_merge() */
    
	req->pub.data = req->pub.user;
	req->pub.data_len = nh->data_len;
	req->pub.xheader_len = nh->xheader_len;
	req->partner_req = NULL;

	D_TR(printf("pscom_generate_recv_req(). xheaderlen=%d\n", req->pub.xheader_len));

	_genreq_enq(con, req);

	return req;
}


static
pscom_req_t *pscom_get_default_recv_req(pscom_con_t *con, pscom_header_net_t *nh)
{
	pscom_request_t *(*default_recv)(pscom_connection_t *connection,
					 pscom_header_net_t *header_net)
		= get_sock(con->pub.socket)->pub.ops.default_recv;
	if (default_recv) {
		pscom_request_t *ureq;
		pscom_req_t *req;

		ureq = default_recv(&con->pub, nh);
		if (ureq) {
			assert(ureq->state & PSCOM_REQ_STATE_DONE);
			ureq->connection = &con->pub;
			ureq->socket = con->pub.socket;
			req = get_req(ureq);
			assert(req->magic == MAGIC_REQUEST);
		} else {
			req = NULL;
		}

		return req;
	} else {
		return NULL;
	}
}


static inline
pscom_req_t *_pscom_get_user_receiver(pscom_con_t *con, pscom_header_net_t *nh)
{
	pscom_req_t *req;
	req = pscom_get_default_recv_req(con, nh);
	if (!req) {
		req = _pscom_get_posted_recv_req(con, nh);
		
		if (!req) {
			req = _pscom_generate_recv_req(con, nh);
			assert(req);
		}
	}
	return req;
}


static
pscom_req_t *pscom_get_rma_write_receiver(pscom_con_t *con, pscom_header_net_t *nh)
{
	pscom_req_t *req;
	pscom_xheader_rma_write_t *rma_header = &nh->xheader->rma_write;
	
	req = pscom_req_create(0, 0);
	req->pub.state = PSCOM_REQ_STATE_RMA_WRITE_REQUEST | PSCOM_REQ_STATE_PASSIVE_SIDE;
	
	/* freed in io_done() */
	req->pub.data = rma_header->dest;
	req->pub.data_len = nh->data_len;
	req->pub.xheader_len = 0;
	req->pub.ops.io_done = pscom_request_free;
	
	D_TR(printf("pscom_get_rma_write_receiver(). dest=%p, len=%d\n",
		    req->pub.data, req->pub.data_len));

	return req;
}


static
pscom_req_t *_pscom_get_rma_read_receiver(pscom_con_t *con, pscom_header_net_t *nh)
{
	pscom_xheader_rma_read_t *rma_header = &nh->xheader->rma_read;

	pscom_xheader_rma_read_answer_t rma_answer;
		
	rma_answer.id = rma_header->id;

	_pscom_send_inplace(con, PSCOM_MSGTYPE_RMA_READ_ANSWER,
			    &rma_answer, sizeof(rma_answer),
			    rma_header->src, rma_header->src_len,
			    NULL, 0);
			    
	return NULL;
}


static
pscom_req_t *_pscom_get_rma_read_answer_receiver(pscom_con_t *con, pscom_header_net_t *nh)
{
	pscom_req_t *req;
	pscom_xheader_rma_read_answer_t *rma_answer;
	pscom_xheader_rma_read_t *rma_request;

	assert(!list_empty(con->rma_readq.next));
	
	req = list_entry(con->rma_readq.next, pscom_req_t, next);

	rma_answer = &nh->xheader->rma_read_answer;
	rma_request = &req->pub.xheader.rma_read;
	
	assert(rma_answer->id == rma_request->id);

	_pscom_rma_read_req_deq(req);

	return req;
}


static inline
void pscom_post_send_direct(pscom_req_t *req, unsigned msg_type);

	
static
void pscom_rendezvous_read_data_io_done(pscom_request_t *request)
{
	pscom_req_t *req = get_req(request);
	pscom_req_t *user_req = req->partner_req;

	pscom_rendezvous_data_t *rd =
		(pscom_rendezvous_data_t *) req->pub.user;
	
	pscom_recv_req_done(user_req);

	/* rewrite rendezvous_req for rendezvous fin message */
	req->pub.xheader.ren_fin.id = rd->msg.id;
	req->pub.xheader_len = sizeof(req->pub.xheader.ren_fin);
		
	req->pub.data = NULL;
	req->pub.data_len = 0;
	
	/* rendezvous_req->pub.connection already set */
	req->pub.ops.io_done = pscom_request_free;

	pscom_post_send_direct(req, PSCOM_MSGTYPE_RENDEZVOUS_FIN);
}


static inline
void _pscom_post_rma_read(pscom_req_t *req);


static
void _pscom_rendezvous_read_data(pscom_req_t *user_recv_req, pscom_req_t *rendezvous_req)
{
	pscom_rendezvous_data_t *rd =
		(pscom_rendezvous_data_t *) rendezvous_req->pub.user;
	
	unsigned int to_read = pscom_min(rd->msg.data_len, user_recv_req->pub.data_len);
	pscom_con_t *con = get_con(rendezvous_req->pub.connection);

	/* rewrite the rendezvous_req for read rma (read data) */
	rendezvous_req->pub.data_len = to_read;
	rendezvous_req->pub.data = user_recv_req->pub.data;

	/* rendezvous_req->pub.connection already set */
	rendezvous_req->pub.xheader.rma_read.src = rd->msg.data;
	
	rendezvous_req->pub.ops.io_done = pscom_rendezvous_read_data_io_done;
	rendezvous_req->partner_req = user_recv_req;
	
	if (rd->use_arch_read && con->rma_read) {
//#define RMA_CNT
#ifdef RMA_CNT
		static unsigned work_cnt = 0;
		static unsigned fail_cnt = 0;
#endif
		if (con->rma_read(rendezvous_req, rd))  {
#ifdef RMA_CNT
			fail_cnt++;
			if (fail_cnt % 1000 == 0) {
				printf("WorkCnt:%u, FailCnt: %u\n", work_cnt, fail_cnt);
			}
#endif
			goto rma_read_fallback;
		}
#ifdef RMA_CNT
		work_cnt++;
		if (work_cnt % 1000 == 0) {
			printf("WorkCnt:%u, FailCnt: %u\n", work_cnt, fail_cnt);
		}
#endif		
	} else {
	rma_read_fallback:
		_pscom_post_rma_read(rendezvous_req);
	}
}


static
void pscom_rendezvous_receiver_io_done(pscom_request_t *req)
{
	pscom_rendezvous_data_t *rd =
		(pscom_rendezvous_data_t *) req->user;

	/* rewrite the header */
	req->header.msg_type = PSCOM_MSGTYPE_USER;
	/* req->header.xheader_len already set */
	req->header.data_len = rd->msg.data_len;

	pscom_lock(); {
		/* Use the rewritten header to search for a recv request: */
		pscom_req_t *user_req = _pscom_get_user_receiver(get_con(req->connection),
								 &req->header);
		assert(user_req);
		pscom_req_prepare_recv(user_req, &req->header, req->connection);

		if (!(user_req->pub.state & PSCOM_REQ_STATE_GRECV_REQUEST)) {
			/* found user receive request. Initiate a
			   rma_read. */
			_pscom_rendezvous_read_data(user_req, get_req(req));
		} else {
			/* found generated request.
			   Continue after user post a recv. */
			user_req->partner_req = get_req(req);
		}
	} pscom_unlock();
}


static
pscom_req_t *pscom_get_rendezvous_receiver(pscom_con_t *con, pscom_header_net_t *nh)
{
	pscom_req_t *req;

	req = pscom_req_create(nh->xheader_len, sizeof(pscom_rendezvous_data_t));
	pscom_rendezvous_data_t *rd = (pscom_rendezvous_data_t *) req->pub.user;
	
	req->pub.state = PSCOM_REQ_STATE_RENDEZVOUS_REQUEST | PSCOM_REQ_STATE_PASSIVE_SIDE;
	assert(nh->data_len <= sizeof(rd->msg));
	
	req->pub.data = &rd->msg;
	req->pub.data_len = nh->data_len;
	req->pub.xheader_len = nh->xheader_len;
	
	req->pub.ops.io_done = pscom_rendezvous_receiver_io_done;

	rd->use_arch_read = nh->data_len > (sizeof(rd->msg) - sizeof(rd->msg.arch));

	return req;
}


static
pscom_req_t *_pscom_get_rendezvous_fin_receiver(pscom_con_t *con, pscom_header_net_t *nh)
{
	pscom_req_t *user_req = nh->xheader->ren_fin.id;
	pscom_req_t *req = user_req->partner_req;
	pscom_rendezvous_data_t *rd = (pscom_rendezvous_data_t *)req->pub.user;

	if (con->rma_mem_deregister) {
		con->rma_mem_deregister(con, rd);
	}

	pscom_request_free(&req->pub);
	
	_pscom_send_req_done(user_req); // done

	return NULL;
}


/* return a request, which will receive this message.
   return NULL if this message should be discarded */
static
pscom_req_t *_pscom_get_recv_req(pscom_con_t *con, pscom_header_net_t *nh)
{
	pscom_req_t *req;
	D_TR(printf("%s(con:%p, nh->msg_type:%u)\n",
		    __func__, con, nh->msg_type));

	if (nh->msg_type == PSCOM_MSGTYPE_USER) {
		req = _pscom_get_user_receiver(con, nh);
		pscom_req_prepare_recv(req, nh, &con->pub);
	} else {
		switch (nh->msg_type) {
		case PSCOM_MSGTYPE_RMA_WRITE:
			req = pscom_get_rma_write_receiver(con, nh);
			break;
		case PSCOM_MSGTYPE_RMA_READ:
			req = _pscom_get_rma_read_receiver(con, nh);
			break;
		case PSCOM_MSGTYPE_RMA_READ_ANSWER:
			req = _pscom_get_rma_read_answer_receiver(con, nh);
			break;
		case PSCOM_MSGTYPE_RENDEZVOUS_REQ:
			req = pscom_get_rendezvous_receiver(con, nh);
			break;
		case PSCOM_MSGTYPE_RENDEZVOUS_FIN:
			req = _pscom_get_rendezvous_fin_receiver(con, nh);
			break;
		default:
			DPRINT(1, "Receive unknown msg_type %u", nh->msg_type);
			req = NULL;
		}
		if (req) pscom_req_prepare_recv(req, nh, &con->pub);
	}
	
	return req;
}


// return true at the end of each message
int pscom_read_is_at_message_start(pscom_con_t *con)
{
	return !con->in.req && !con->in.skip;
}


void
pscom_read_get_buf(pscom_con_t *con, char **buf, size_t *len)
{
	if (con->in.req) {
		pscom_req_t *req = con->in.req;
		*buf = req->cur_data.iov_base;
		*len = req->cur_data.iov_len;
		assert(req->cur_data.iov_len > 0);
	} else if (!con->in.skip) {
		unsigned int readlen = pscom.env.readahead;

		if (con->in.readahead.iov_len >= sizeof(pscom_header_net_t)) {
			readlen = header_length((pscom_header_net_t *)con->in.readahead.iov_base);
		}
		_check_readahead(con, readlen);

		*buf = con->in.readahead.iov_base + con->in.readahead.iov_len;
		*len = readlen - con->in.readahead.iov_len;
	} else {
		unsigned int rlen = pscom_min(pscom.env.skipblocksize, con->in.skip);
		_check_readahead(con, rlen);
		*buf = con->in.readahead.iov_base;
		*len = rlen;
	}
	
	D_TR(printf("pscom_read_get_buf(con, *buf=%p, *len=%zu)\n",
		    *buf, *len));
}


void
pscom_read_done(pscom_con_t *con, char *buf, size_t len)
{
	pscom_req_t *req = con->in.req;
	
	D_TR(printf("pscom_read_done(con, buf=%p, len=%zu, %s)\n",
		    buf, len, pscom_dumpstr(buf, pscom_min(len, 32))));

	if (!len) goto err_eof;

	if (req) {
		unsigned int _len;
		
		_len = req_write(req, buf, len);
		len -= _len;
		buf += _len;

		_pscom_update_recvq(con);

		assert(!con->in.readahead.iov_len);
	
		if (!len) return;

		assert(!con->in.req);
	}

	if (con->in.readahead.iov_len) {
		char *dest;
		// append buf,len to readahead buffer
		assert(!con->in.skip);

		_check_readahead(con, con->in.readahead.iov_len + len);
		dest = ((char *)con->in.readahead.iov_base) + con->in.readahead.iov_len;
		if (buf != dest) {
			memcpy(dest, buf, len);
		}
		
		con->in.readahead.iov_len += len;

		buf = con->in.readahead.iov_base;
		len = con->in.readahead.iov_len;
	} else if (con->in.skip) {
		if (con->in.skip < len) {
			buf += con->in.skip;
			len -= con->in.skip;
			con->in.skip = 0;
		} else {
			con->in.skip -= len;
			return;
		}
	}


	while (header_complete(buf, len)) {
		// consume data
		pscom_header_net_t *header = (pscom_header_net_t *)buf;
		unsigned int hlen = header_length(header);
		unsigned int l;
		
		con->in.req = _pscom_get_recv_req(con, header);
		req = con->in.req;

		buf += hlen;
		len -= hlen;

		if (req) {
			req->pub.state |= PSCOM_REQ_STATE_IO_STARTED;
			l = req_write(req, buf, len);
			buf += l;
			len -= l;
		} else {
			/* Skip message */
			unsigned skip = pscom_min(header->data_len, len);

			buf += skip;
			len -= skip;
			con->in.skip = header->data_len - skip;
		}

		_pscom_update_recvq(con);
		assert(!con->in.skip || !len);
	}

	
	if (len && (con->in.readahead.iov_base != buf)) {
		// save unused data
		_check_readahead(con, len);
		memmove(con->in.readahead.iov_base, buf, len);
	}
	con->in.readahead.iov_len = len;
	
	return;
	/* --- */
err_eof:
	pscom_con_error(con, PSCOM_OP_READ, PSCOM_ERR_EOF);
	return;
}


pscom_req_t *pscom_write_get_iov(pscom_con_t *con, struct iovec iov[2])
{
	if (!list_empty(&con->sendq)) {
		pscom_req_t *req = list_entry(con->sendq.next, pscom_req_t, next);

		iov[0] = req->cur_header;
		iov[1] = req->cur_data;
		
		req->pub.state |= PSCOM_REQ_STATE_IO_STARTED;

		return req;
	} else {
		con->write_stop(con);
		return 0;
	}
}


void pscom_write_done(pscom_con_t *con, pscom_req_t *req, size_t len)
{
	pscom_forward_iov(&req->cur_header, len);

	if (!(req->cur_header.iov_len + req->cur_data.iov_len)) {
		_pscom_send_req_deq(con, req);

		_pscom_send_req_done(req); // done
	}
}


static
void _pscom_send(pscom_con_t *con, unsigned msg_type,
		 void *xheader, unsigned xheader_len,
		 void *data, unsigned data_len)
{
	pscom_req_t *req;

	req = pscom_req_create(xheader_len, data_len);

	req->pub.xheader_len = xheader_len;
	req->pub.data_len = data_len;
	req->pub.data = req->pub.user;

	memcpy(&req->pub.xheader, xheader, xheader_len);
	memcpy(req->pub.data, data, data_len);

	req->pub.connection = &con->pub;
	req->pub.ops.io_done = pscom_request_free;
	
	pscom_req_prepare_send(req, msg_type);
	
	req->pub.state = PSCOM_REQ_STATE_SEND_REQUEST | PSCOM_REQ_STATE_POSTED;
		
	_pscom_send_req_enq(con, req);
}


struct pscom_req_send_inplace_rdata
{
	void (*io_done)(pscom_req_state_t state, void *priv);
	void *priv;
	char data[0];
};


static
void pscom_send_inplace_io_done(pscom_request_t *req)
{
	struct pscom_req_send_inplace_rdata *rdata =
		(struct pscom_req_send_inplace_rdata *)req->user;

	if (rdata->io_done) {
		rdata->io_done(req->state, rdata->priv);
	}
	
	pscom_request_free(req);
}


void _pscom_send_inplace(pscom_con_t *con, unsigned msg_type,
			 void *xheader, unsigned xheader_len,
			 void *data, unsigned data_len,
			 void (*io_done)(pscom_req_state_t state, void *priv), void *priv)
{
	pscom_req_t *req;

	struct pscom_req_send_inplace_rdata *rdata;

	req = pscom_req_create(xheader_len, sizeof(*rdata));

	req->pub.xheader_len = xheader_len;
	req->pub.data_len = data_len;
	req->pub.data = data;
	rdata = (struct pscom_req_send_inplace_rdata *)req->pub.user;

	rdata->io_done = io_done;
	rdata->priv = priv;
	
	memcpy(&req->pub.xheader, xheader, xheader_len);
		
	req->pub.connection = &con->pub;
	req->pub.ops.io_done = pscom_send_inplace_io_done;
	
	pscom_req_prepare_send(req, msg_type);
	
	req->pub.state = PSCOM_REQ_STATE_SEND_REQUEST | PSCOM_REQ_STATE_POSTED;
		
	_pscom_send_req_enq(con, req);
}


static
int _pscom_cancel_send(pscom_req_t *req)
{
	if (req->pub.state & PSCOM_REQ_STATE_IO_DONE) {
		return 0;
	}
	if (req->pub.state & PSCOM_REQ_STATE_IO_STARTED) {
		return 0;
	}

	_pscom_send_req_deq(get_con(req->pub.connection), req);

	req->pub.state |= PSCOM_REQ_STATE_CANCELED;
	_pscom_send_req_done(req); // done

	return 1;
}


static
int _pscom_cancel_recv(pscom_req_t *req)
{
	if (req->pub.state & PSCOM_REQ_STATE_IO_DONE) {
		return 0;
	}
	if (req->pub.state & PSCOM_REQ_STATE_IO_STARTED) {
		return 0;
	}

	assert(_is_in_recv_queue(req));
	
	_pscom_recv_req_deq(req);
	_pscom_recvq_any_cleanup(get_sock(req->pub.socket));
	
	req->pub.state |= PSCOM_REQ_STATE_CANCELED;
	_pscom_recv_req_done(req); // done

	return 1;
}


static inline
void pscom_post_send_direct(pscom_req_t *req, unsigned msg_type)
{
	pscom_req_prepare_send(req, msg_type); // build header and iovec

	D_TR(printf("%s(req:%p,%s)\n", __func__, req, pscom_req_state_str(req->pub.state)));

	pscom_lock(); {
		req->pub.state = PSCOM_REQ_STATE_SEND_REQUEST | PSCOM_REQ_STATE_POSTED;
		
		_pscom_send_req_enq(get_con(req->pub.connection), req);
	} pscom_unlock();
}


static inline
void pscom_post_send_rendezvous(pscom_req_t *user_req)
{

	pscom_req_t *req;
	pscom_rendezvous_data_t *rd;
	
	req = pscom_req_create(user_req->pub.xheader_len,
			       sizeof(pscom_rendezvous_data_t));

	req->pub.xheader_len = user_req->pub.xheader_len;
	req->pub.data_len = sizeof(rd->msg) - sizeof(rd->msg.arch);
	req->pub.data = req->pub.user;

	rd = (pscom_rendezvous_data_t *)req->pub.data;

	rd->msg.id = user_req;
	rd->msg.data = user_req->pub.data;
	rd->msg.data_len = user_req->pub.data_len;

	pscom_con_t *con = get_con(user_req->pub.connection);
	
	if (con->rma_mem_register) {
		req->pub.data_len += con->rma_mem_register(con, rd);
	}
	
	memcpy(&req->pub.xheader, &user_req->pub.xheader, user_req->pub.xheader_len);
	
	req->pub.connection = user_req->pub.connection;
	req->pub.ops.io_done = NULL;

	user_req->partner_req = req;
	user_req->pub.state = PSCOM_REQ_STATE_RENDEZVOUS_REQUEST |
		PSCOM_REQ_STATE_SEND_REQUEST | PSCOM_REQ_STATE_POSTED;

	pscom_post_send_direct(req, PSCOM_MSGTYPE_RENDEZVOUS_REQ);
}


static inline
void _pscom_post_rma_read(pscom_req_t *req)
{
	pscom_xheader_rma_read_t xheader;
	
	req->pub.xheader.rma_read.id = req;
	
	req->pub.state = PSCOM_REQ_STATE_RMA_READ_REQUEST | PSCOM_REQ_STATE_POSTED;
	_pscom_rma_read_req_enq(req);
	
	xheader.id = req;
	xheader.src = req->pub.xheader.rma_read.src;
	xheader.src_len = req->pub.data_len;
	
	_pscom_send(get_con(req->pub.connection),
		    PSCOM_MSGTYPE_RMA_READ, &xheader, sizeof(xheader), NULL, 0);
}


/*
******************************************************************************
*/

pscom_request_t *pscom_request_create(unsigned int max_xheader_len, unsigned int user_size)
{
	pscom_req_t *req;

	req = pscom_req_create(max_xheader_len, user_size);

	return req ? &req->pub : NULL;
}


void pscom_request_free(pscom_request_t *request)
{
	pscom_req_free(get_req(request));
}


void pscom_post_recv(pscom_request_t *request)
{
	pscom_req_t *req = get_req(request);
	assert(req->magic == MAGIC_REQUEST);
	assert(request->state & PSCOM_REQ_STATE_DONE);
	assert((request->connection != NULL) || (request->socket != NULL));

	D_TR(printf("%s(req:%p,%s)\n", __func__, req, pscom_req_state_str(req->pub.state)));
	
	pscom_lock(); {
		pscom_req_t *genreq;
		
		req->pub.state = PSCOM_REQ_STATE_RECV_REQUEST | PSCOM_REQ_STATE_POSTED;

		genreq = _genreq_find(req);

		if (!genreq) {
			_pscom_recv_req_enq(req);
		} else {
			_genreq_dequeue(genreq);
			_genreq_merge(req, genreq);
		}
	} pscom_unlock();
}


/* return 1, if there is a matching receive. 0 otherwise. */
/* in case 1: copy also the message header */
int pscom_iprobe(pscom_request_t *request)
{
	pscom_req_t *req = get_req(request);
	int res;
	assert(req->magic == MAGIC_REQUEST);
	assert(request->state & PSCOM_REQ_STATE_DONE);
	assert((request->connection != NULL) || (request->socket != NULL));

	D_TR(printf("%s(req:%p,%s)\n", __func__, req, pscom_req_state_str(req->pub.state)));
	
	pscom_lock(); {
		pscom_req_t *genreq;
		
		req->pub.state = PSCOM_REQ_STATE_RECV_REQUEST | PSCOM_REQ_STATE_POSTED;

		pscom_progress(0);
		genreq = _genreq_find(req);

		if (!genreq || (!(genreq->pub.state & PSCOM_REQ_STATE_DONE))) {
			res = 0;
		} else {
			res = 1;

			genreq_merge_header(req, genreq);
		}
		req->pub.state |= PSCOM_REQ_STATE_DONE;
	} pscom_unlock();

	return res;
}

void pscom_probe(pscom_request_t *request)
{
	while (!pscom_iprobe(request)) {
		pscom_wait_any();
	}
}


void pscom_post_send(pscom_request_t *request)
{
	pscom_req_t *req = get_req(request);
	assert(req->magic == MAGIC_REQUEST);
	assert(request->state & PSCOM_REQ_STATE_DONE);
	assert(request->connection != NULL);

	if (req->pub.data_len < get_con(request->connection)->rendezvous_size) {
		// printf("Send s %d\n", request->data_len);
		pscom_post_send_direct(req, PSCOM_MSGTYPE_USER);
	} else {
		// printf("Send r %d\n", request->data_len);
		pscom_post_send_rendezvous(req);
	}
}
		

void pscom_send(pscom_connection_t *connection,
		void *xheader, unsigned int xheader_len,
		void *data, unsigned int data_len)
{
	pscom_lock(); {
		_pscom_send(get_con(connection), PSCOM_MSGTYPE_USER,
			    xheader, xheader_len,
			    data, data_len);
	} pscom_unlock();
}


void pscom_send_inplace(pscom_connection_t *connection,
			void *xheader, unsigned int xheader_len,
			void *data, unsigned int data_len,
			void (*io_done)(pscom_req_state_t state, void *priv), void *priv)
{
	pscom_lock(); {
		_pscom_send_inplace(get_con(connection), PSCOM_MSGTYPE_USER,
				    xheader, xheader_len,
				    data, data_len,
				    io_done, priv);
	} pscom_unlock();
}


pscom_err_t pscom_recv(pscom_connection_t *connection, pscom_socket_t *socket,
		       void *xheader, unsigned int xheader_len,
		       void *data, unsigned int data_len)
{
	pscom_request_t *req = pscom_request_create(xheader_len, 0);
	pscom_err_t ret = PSCOM_ERR_IOERROR;
	
	if (!req) {
		return PSCOM_ERR_STDERROR;
	}
	
	req->xheader_len = xheader_len;
	// memcpy(req->xheader.user, xheader, xheader_len);
	req->data_len = data_len;
	req->data = data;
	req->connection = connection;
	req->socket = socket;

	pscom_post_recv(req);

	pscom_wait(req);

	if (pscom_req_successful(req)) {
		memcpy(xheader, req->xheader.user, xheader_len);
		ret = PSCOM_SUCCESS;
	}

	pscom_request_free(req);

	return ret;
}


void pscom_post_rma_write(pscom_request_t *request)
{
	pscom_req_t *req = get_req(request);
	assert(req->magic == MAGIC_REQUEST);
	assert(request->state & PSCOM_REQ_STATE_DONE);
	assert(request->connection != NULL);

	pscom_req_prepare_rma_write(req); // build header and iovec

	pscom_post_send_direct(req, PSCOM_MSGTYPE_RMA_WRITE);
}


void pscom_post_rma_read(pscom_request_t *request)
{
	pscom_req_t *req = get_req(request);
	assert(req->magic == MAGIC_REQUEST);
	assert(request->state & PSCOM_REQ_STATE_DONE);
	assert(request->connection != NULL);

	D_TR(printf("%s(req:%p,%s)\n", __func__, req, pscom_req_state_str(req->pub.state)));

	pscom_lock(); {
		_pscom_post_rma_read(req);
	} pscom_unlock();
}


void pscom_wait(pscom_request_t *request)
{
	volatile pscom_req_state_t *state = &request->state;
	
	while (!(*state & PSCOM_REQ_STATE_DONE)) {
		pscom_wait_any();
	}
}


int pscom_cancel_send(pscom_request_t *request)
{
	pscom_req_t *req = get_req(request);
	int res;
	D_TR(printf("%s\n", __func__));
	assert(req->magic == MAGIC_REQUEST);
	assert(request->state & PSCOM_REQ_STATE_SEND_REQUEST);
	if (request->state & PSCOM_REQ_STATE_DONE) return 0;

	pscom_lock(); {
		res = _pscom_cancel_send(req);
	} pscom_unlock();

	return res;
}


int pscom_cancel_recv(pscom_request_t *request)
{
	pscom_req_t *req = get_req(request);
	int res;
	D_TR(printf("%s\n", __func__));
	assert(req->magic == MAGIC_REQUEST);
	assert(request->state & PSCOM_REQ_STATE_RECV_REQUEST);
	if (request->state & PSCOM_REQ_STATE_DONE) return 0;

	pscom_lock(); {
		res = _pscom_cancel_recv(req);
	} pscom_unlock();

	return res;
}


int pscom_cancel(pscom_request_t *request)
{
	pscom_req_t *req = get_req(request);
	assert(req->magic == MAGIC_REQUEST);

	if (request->state & PSCOM_REQ_STATE_SEND_REQUEST) {
		return pscom_cancel_send(request);
	} else if (request->state & PSCOM_REQ_STATE_RECV_REQUEST) {
		return pscom_cancel_recv(request);
	}
	return 0;
}
