/**************************************************** -*- linux-c -*- *
 *                  ParaStation
 *
 * Copyright (c) 2007 ParTec Cluster Competence Center GmbH
 * All rights reserved.
 **********************************************************************/
/**
 * $Id: pscom_priv.h 5310 2007-10-24 15:26:35Z hauke $
 *
 * @author
 *         Jens Hauke <hauke@par-tec.de>
 *
 **********************************************************************/
#ifndef _PSCOM_PRIV_H_
#define _PSCOM_PRIV_H_

#include "pscom.h"
#include "pscom_types.h"
#include "list.h"
#include "pscom_ufd.h"
#include <string.h>
#include <pthread.h>
#include <unistd.h>
#include <assert.h>
#include <sys/types.h>

#include "pscom_tcp.h"
#include "pscom_shm.h"
#include "pscom_p4s.h"
#include "pscom_gm.h"

#include "pscom_debug.h"


#define MAGIC_REQUEST		0x72657175
struct PSCOM_req
{
	unsigned long magic;
	struct list_head next; // general purpose next (recvq, sendq ...)
	struct list_head gen_next_any; // used by list pscom_sock_t.genrecvq_any
	struct list_head all_req_next; // used by list struct PSCOM.requests
	
	struct iovec cur_header;
	struct iovec cur_data;
	unsigned int skip;

	/* partner_req:
	   rma send:
	   - user req point to rendezvous_req (PSCOM_MSGTYPE_RENDEZVOUS_REQ).
	   rma recv:
	   - generated request point to rendezvous_req
	   - rendezvous requests point to user recv request.
	*/
	pscom_req_t *partner_req;
	
	
	unsigned int req_no; // debug counter
	pscom_request_t pub;
};


typedef struct psib_conn {
	void	*priv;
} psib_conn_t;


typedef struct psoib_conn {
	struct psoib_con_info *mcon;
} psoib_conn_t;


typedef struct psdapl_conn {
	struct psdapl_con_info *ci;
} psdapl_conn_t;


typedef struct pselan_conn {
	struct pselan_con_info *ci;
} pselan_conn_t;


typedef struct user_conn {
	void	*priv;
} user_conn_t;


/* rendezvous message for RMA requests. */
typedef struct pscom_rendezvous_msg {
	void		*id; /* == pscom_req_t *user_req; from sending side */
	void		*data;
	unsigned int	data_len;
	union {
		struct {} shm;
		struct {
			uint32_t /* DAT_RMR_CONTEXT */	rmr_context;
			uint64_t /* DAT_CONTEXT */	rmr_vaddr;
		} dapl;
	}	arch;
} pscom_rendezvous_msg_t;

typedef struct pscom_rendezvous_data_shm {
} pscom_rendezvous_data_shm_t;


typedef struct _pscom_rendezvous_data_dapl {
	char /* struct psdapl_rdma_req */ data[128];
} _pscom_rendezvous_data_dapl_t;


typedef struct pscom_rendezvous_data {
	pscom_rendezvous_msg_t	msg;
	int	use_arch_read;
	union {
		pscom_rendezvous_data_shm_t	shm;
		_pscom_rendezvous_data_dapl_t	dapl;
	}		arch;
} pscom_rendezvous_data_t;


#define MAGIC_CONNECTION	0x78626c61
struct PSCOM_con
{
	unsigned long		magic;
	struct list_head	next;
	void (*read_start)(pscom_con_t *con);
	void (*read_stop)(pscom_con_t *con);
	void (*write_start)(pscom_con_t *con);
	void (*write_stop)(pscom_con_t *con);
	void (*do_write)(pscom_con_t *con);
	void (*close)(pscom_con_t *con);

	/* RMA functions: */
	/* register mem region for RMA. should return size of
	 * rd->msg.arch.xxx (this is used, to calculate the size of
	 * the rendezvous message). return 0 to disable arch read (in
	 * case of a failure). */
	unsigned int (*rma_mem_register)(pscom_con_t *con, pscom_rendezvous_data_t *rd);
	/* deregister mem. */
	void (*rma_mem_deregister)(pscom_con_t *con, pscom_rendezvous_data_t *rd);
	/* return -1 on error.
	   see _pscom_rendezvous_read_data()  */
	int (*rma_read)(pscom_req_t *rendezvous_req, pscom_rendezvous_data_t *rd);
	
	unsigned int		rendezvous_size;
	
	struct list_head	sendq;		// List of pscom_req_t.next
	struct list_head	recvq;		// List of pscom_req_t.next

	struct list_head	genrecvq;	// List of pscom_req_t.next
	struct list_head	rma_readq;	// List of pscom_req_t.next

	pscom_poll_reader_t	poll_reader;
	struct list_head	poll_next_send; // used by pscom.poll_sender

	struct {
		pscom_req_t	*req;

		struct iovec	readahead;
		unsigned int	readahead_size;

		unsigned int	skip;
	}			in;
	
	union {
		tcp_conn_t	tcp;
		shm_conn_t	shm;
		p4s_conn_t	p4s;
		psib_conn_t	mvapi;
		psoib_conn_t	openib;
		psgm_conn_t	gm;
		psdapl_conn_t	dapl;
		pselan_conn_t	elan;
		user_conn_t	user; // Future usage (new plugins)
	}			arch;

	pscom_connection_t	pub;
};


#define MAGIC_SOCKET		0x6a656e73
struct PSCOM_sock
{
	unsigned long		magic;
	struct list_head	next;		// used by list pscom.sockets
	
	struct list_head	connections;	// List of pscom_con_t.next

	struct list_head	recvq_any;	// List of pscom_req_t.next (all recv any requests)
	struct list_head	genrecvq_any;	// List of pscom_req_t.gen_next_any(all generated requests)

	int			listen_fd;

	tcp_sock_t		tcp;
	shm_sock_t		shm;
	p4s_sock_t		p4s;
//	psib_sock_t		mvapi;
//	psoib_sock_t		openib;
	psgm_sock_t		gm;
//	psdapl_sock_t		dapl;
//	pselan_sock_t		elan;

	pscom_socket_t		pub;
};


struct PSCOM
{
	ufd_t			ufd;
	struct list_head	sockets; // List of pscom_sock_t.next
	struct list_head	requests; // List of pscom_req_t.all_req_next
	
	pthread_mutex_t		global_lock;
	pthread_mutex_t		lock_requests;

	struct list_head	io_doneq; // List of pscom_req_t.next

	struct list_head	poll_reader;	// List of pscom_poll_reader_t.next
	struct list_head	poll_sender;	// List of pscom_con_t.poll_next_send
	
	struct {
		int		debug;
		unsigned int	so_sndbuf;
		unsigned int	so_rcvbuf;
		int		tcp_nodelay;
		int		sched_yield;
		unsigned int	rendezvous_size;
		unsigned int	rendezvous_size_shm;
		unsigned int	rendezvous_size_dapl;
		unsigned int	rendezvous_size_elan;
		int		sigquit;
		unsigned int	readahead;
		unsigned int	retry;
		unsigned int	skipblocksize;

		char		*plugindir;
	}			env;

	struct {
		unsigned int	reqs;
		unsigned int	gen_reqs;
		unsigned int	gen_reqs_used;

		unsigned int	progresscounter;
		unsigned int	progresscounter_check;
	}			stat;
};

extern pscom_t pscom;

#define PSCOM_ARCH_ERROR	101
#define PSCOM_ARCH_LOOP		102
#define PSCOM_ARCH_TCP		103
#define PSCOM_ARCH_SHM		104
#define PSCOM_ARCH_P4S		105
#define PSCOM_ARCH_GM		106
#define PSCOM_ARCH_MVAPI	107
#define PSCOM_ARCH_OPENIB	108
#define PSCOM_ARCH_ELAN		109
#define PSCOM_ARCH_DAPL		110

#define PSCOM_TCP_PRIO		2
#define PSCOM_SHM_PRIO		90
#define PSCOM_P4S_PRIO		10
#define PSCOM_GM_PRIO		15
#define PSCOM_MVAPI_PRIO	20
#define PSCOM_OPENIB_PRIO	20
#define PSCOM_ELAN_PRIO		20
#define PSCOM_DAPL_PRIO		15


#define PSCOM_MSGTYPE_USER	0
#define PSCOM_MSGTYPE_RMA_WRITE	1
#define PSCOM_MSGTYPE_RMA_READ	2
#define PSCOM_MSGTYPE_RMA_READ_ANSWER	3
#define PSCOM_MSGTYPE_RENDEZVOUS_REQ	4 /* Request for a rendezvous */
#define PSCOM_MSGTYPE_RENDEZVOUS_FIN	5 /* Rendezvous done */


static inline
void pscom_lock(void)
{
	int res_mutex_lock;
	res_mutex_lock = pthread_mutex_lock(&pscom.global_lock);
	assert(res_mutex_lock == 0);
}


void pscom_unlock(void);

static inline
void _pscom_unlock(void)
{
	int res_mutex_unlock;
	res_mutex_unlock = pthread_mutex_unlock(&pscom.global_lock);
	assert(res_mutex_unlock == 0);
}


static inline
void pscom_call_io_done(void)
{
	pscom_unlock();
	pscom_lock();
}


static inline
pscom_con_t *get_con(pscom_connection_t *con)
{
	return list_entry(con, pscom_con_t, pub);
}


static inline
pscom_sock_t *get_sock(pscom_socket_t *socket)
{
	return list_entry(socket, pscom_sock_t, pub);
}


static inline
pscom_req_t *get_req(pscom_request_t *request)
{
	return list_entry(request, pscom_req_t, pub);
}


void
pscom_read_get_buf(pscom_con_t *con, char **buf, size_t *len);

void
pscom_read_done(pscom_con_t *con, char *buf, size_t len);

// return true at the end of each message (no current request)
int pscom_read_is_at_message_start(pscom_con_t *con);

pscom_req_t *pscom_write_get_iov(pscom_con_t *con, struct iovec iov[2]);
void pscom_write_done(pscom_con_t *con, pscom_req_t *req, size_t len);


void pscom_con_error(pscom_con_t *con, pscom_op_t operation, pscom_err_t error);

/*
void _pscom_send(pscom_con_t *con, unsigned msg_type,
		 void *xheader, unsigned xheader_len,
		 void *data, unsigned data_len);
*/

void _pscom_send_inplace(pscom_con_t *con, unsigned msg_type,
			 void *xheader, unsigned xheader_len,
			 void *data, unsigned data_len,
			 void (*io_done)(pscom_req_state_t state, void *priv), void *priv);

void pscom_poll_write_start(pscom_con_t *con);
void pscom_poll_write_stop(pscom_con_t *con);
void pscom_poll_read_start(pscom_con_t *con);
void pscom_poll_read_stop(pscom_con_t *con);

int pscom_progress(int timeout);
#endif /* _PSCOM_PRIV_H_ */
