/**************************************************** -*- linux-c -*- *
 *                  ParaStation
 *
 * Copyright (c) 2007 ParTec Cluster Competence Center GmbH
 * All rights reserved.
 **********************************************************************/
/**
 * $Id: pscom_sock.c 5292 2007-10-18 13:09:21Z hauke $
 *
 * @author
 *         Jens Hauke <hauke@par-tec.de>
 *
 **********************************************************************/
#include "pscom_sock.h"
#include "pscom_con.h"
#include "pscom_io.h"
#include <stdlib.h>

static
void _pscom_sock_terminate_all_recvs(pscom_sock_t *sock)
{
	struct list_head *pos, *next;

	// Recvq's of all connections
	list_for_each(pos, &sock->connections) {
		pscom_con_t *con = list_entry(pos, pscom_con_t, next);

		pscom_con_terminate_recvq(con);
	}

	
	// RecvAny Queue:
	list_for_each_safe(pos, next, &sock->recvq_any) {
		pscom_req_t *req = list_entry(pos, pscom_req_t, next);

		list_del(&req->next);
		req->pub.state |= PSCOM_REQ_STATE_ERROR;
		_pscom_recv_req_done(req); // done
	}
}


static
void pscom_sock_stop_listen(pscom_sock_t *sock)
{
	assert(sock->magic == MAGIC_SOCKET);
	
	if (sock->pub.listen_portno == -1)
		return;
		
	ufd_del(&pscom.ufd, sock->listen_fd);
	close(sock->listen_fd);
	sock->listen_fd = -1;
	sock->pub.listen_portno = -1;
}


static
void pscom_sock_close(pscom_sock_t *sock)
{
	assert(sock->magic == MAGIC_SOCKET);

retry:
	pscom_sock_stop_listen(sock);

	while (!list_empty(&sock->connections)) {
		pscom_con_t *con = list_entry(sock->connections.next, pscom_con_t, next);
		pscom_con_close(con);
	}

	_pscom_sock_terminate_all_recvs(sock);
	
	pscom_call_io_done();

	if (!list_empty(&sock->connections) ||
	    !list_empty(&sock->recvq_any) ||
	    sock->pub.listen_portno != -1)
		goto retry; // in the case the io_donq callbacks post more work

	if (!list_empty(&sock->next)) {
		list_del_init(&sock->next);
	}
}


static
void pscom_sock_set_name(pscom_sock_t *sock, char *name)
{
	memset(sock->pub.local_con_info.name, 0, sizeof(sock->pub.local_con_info.name));
	strncpy(sock->pub.local_con_info.name, name, sizeof(sock->pub.local_con_info.name));
}


static
void pscom_sock_init_con_info(pscom_sock_t *sock)
{
	pscom_con_info_t *con_info = &sock->pub.local_con_info;
	char name[20];
	
	con_info->node_id = pscom_get_nodeid();
	con_info->pid = getpid();
	con_info->id = NULL;
	
	snprintf(name, sizeof(name), "p%d", con_info->pid);
	pscom_sock_set_name(sock, name);
}


static
pscom_sock_t *pscom_sock_create(unsigned int userdata_size)
{
	pscom_sock_t *sock;
	sock = malloc(sizeof(*sock) + userdata_size);
	if (!sock) return NULL; // error


	sock->magic = MAGIC_SOCKET;
	sock->pub.ops.con_accept = NULL;
	sock->pub.ops.con_error = NULL;
	sock->pub.ops.default_recv = NULL;

	sock->pub.listen_portno = -1;
	sock->listen_fd = -1;
	sock->pub.userdata_size = userdata_size;
	sock->pub.connection_userdata_size = 0;

	INIT_LIST_HEAD(&sock->connections);
	INIT_LIST_HEAD(&sock->genrecvq_any);
	INIT_LIST_HEAD(&sock->recvq_any);

	pscom_sock_init_con_info(sock);
	
	pscom_plugins_sock_init(sock);

	return sock;
}


static
void pscom_sock_destroy(pscom_sock_t *sock)
{
	assert(sock->magic == MAGIC_SOCKET);
	assert(list_empty(&sock->next));
	assert(list_empty(&sock->connections));
	assert(list_empty(&sock->genrecvq_any));
	assert(list_empty(&sock->recvq_any));

	assert(sock->pub.listen_portno == -1);

	pscom_plugins_sock_destroy(sock);

	sock->magic = 0;
	
	free(sock);
}


/*
******************************************************************************
*/

pscom_socket_t *pscom_open_socket(unsigned int userdata_size,
				  unsigned int connection_userdata_size)
{
	pscom_sock_t *sock;
	
	sock = pscom_sock_create(userdata_size);
	if (!sock) return NULL; // error

	sock->pub.connection_userdata_size = connection_userdata_size;

	pscom_lock(); {
		list_add_tail(&sock->next, &pscom.sockets);
	} pscom_unlock();
	
	return &sock->pub;
}


void pscom_socket_set_name(pscom_socket_t *socket, char *name)
{
	pscom_lock(); {
		pscom_sock_set_name(get_sock(socket), name);
	} pscom_unlock();
}


pscom_err_t pscom_listen(pscom_socket_t *_socket, int portno)
{
	pscom_sock_t *sock = get_sock(_socket);
	pscom_err_t ret = PSCOM_SUCCESS;

	assert(sock->magic == MAGIC_SOCKET);
	
	pscom_lock(); {
		struct sockaddr_in sa;
		unsigned int size;
		int listen_fd;
		int idx;
		
		if (sock->pub.listen_portno != -1)
			goto err_already_listening;

		listen_fd = socket(PF_INET, SOCK_STREAM, 0);
		if (listen_fd < 0) goto err_socket;

		{
			int val = 1;
			setsockopt(listen_fd, SOL_SOCKET, SO_REUSEADDR,
				   (void*) &val, sizeof(val));
		}

		sa.sin_family = AF_INET;
		sa.sin_port = (portno == PSCOM_ANYPORT) ? 0 : htons(portno);
		sa.sin_addr.s_addr = INADDR_ANY;

		if (bind(listen_fd, (struct sockaddr *)&sa, sizeof(sa)) < 0)
			goto err_bind;

		size = sizeof(sa);
		if (getsockname(listen_fd, (struct sockaddr *)&sa, &size) < 0)
			goto err_getsockname;
		
		if (listen(listen_fd, 64) < 0)
			goto err_listen;
    
		sock->pub.listen_portno = ntohs(sa.sin_port);
		sock->listen_fd = listen_fd;

		idx = ufd_add(&pscom.ufd, listen_fd, pscom_con_accept, NULL, NULL, NULL, sock);
		ufd_event_set(&pscom.ufd, idx, POLLIN);

		/* error code */
		if (0) {
		err_listen:
		err_getsockname:
		err_bind:
			close(listen_fd);
		err_socket:
			ret = PSCOM_ERR_STDERROR;
		}
		if (0) {
		err_already_listening:
			ret = PSCOM_ERR_ALREADY;
		}
	} pscom_unlock();

	return ret;
}


void pscom_close_socket(pscom_socket_t *socket)
{
	pscom_lock(); {
		pscom_sock_t *sock = get_sock(socket);
		pscom_sock_close(sock);
		pscom_sock_destroy(sock);
	} pscom_unlock();
}


void pscom_stop_listen(pscom_socket_t *socket)
{
	pscom_lock(); {
		pscom_sock_stop_listen(get_sock(socket));
	} pscom_unlock();
}
