/**************************************************** -*- linux-c -*- *
 *                  ParaStation
 *
 * Copyright (c) 2007 ParTec Cluster Competence Center GmbH
 * All rights reserved.
 **********************************************************************/
/**
 * $Id: pscom_str_util.c 5279 2007-10-15 15:14:11Z hauke $
 *
 * @author
 *         Jens Hauke <hauke@par-tec.de>
 *
 **********************************************************************/
#include "pscom_priv.h"
#include "pscom_str_util.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>


/* Take a service name, and a service type, and return a port number.  If the
   service name is not found, it tries it as a decimal number.  The number
   returned is byte ordered for the network. */
int pscom_atoport(const char *service, const char *proto)
{
	long int lport;
	struct servent *serv;
	char *errpos = NULL;
	
	if (!service) { service = "";}
	
	lport = strtol(service, &errpos, 0);
	if (errpos && *errpos == 0) {
		/* valid integer, or empty string */
		return htons(lport);
	}
	
	if (!proto) {errno = EINVAL; return -1;}
	
	/* Try to read it from /etc/services */
	serv = getservbyname(service, proto);
	
	if (serv != NULL) {
		return serv->s_port;
	} else {
		errno = EINVAL; return -1;
	}
}


int pscom_atoaddr(const char *address, struct in_addr *addr)
{
	struct hostent *mhost;

	if (!addr) {errno = EINVAL; return -1;}

	if (!address) {
		addr->s_addr = INADDR_LOOPBACK;
		return 0;
	}
	
	/* First try it as aaa.bbb.ccc.ddd. */
	if (inet_aton(address, addr)) {
		/* ok */
		return 0; 
	}

	
	/* Get list of IP-addresses */
	mhost = gethostbyname(address);
	/* printf("host %s\n", address);*/
	if (!mhost) goto err;
	if (!mhost->h_addr_list) goto err;

	addr->s_addr = *(in_addr_t *)*mhost->h_addr_list;
	
	return 0;
	/* --- */
err:
	errno = EINVAL; return -1;
}


int pscom_ascii_to_sockaddr_in(const char *host, const char *port,
			       const char *protocol,
			       struct sockaddr_in *addr)
{
	int res;
	int portno;
	struct in_addr inaddr;

	if (!host || !port || !protocol) goto err;
	
	portno = pscom_atoport(port, protocol);
	if (portno < 0) goto err;
	
	res = pscom_atoaddr(host, &inaddr);	
	if (res < 0) goto err;

	if (addr) {
		addr->sin_family = PF_INET;
		addr->sin_port = portno;
		addr->sin_addr = inaddr;
	}

	return 0;
	/* error code */
err:
	return -1;
}


const char *pscom_socket_str(int nodeid, int portno)
{
	static char socket_str[sizeof("xxx.xxx.xxx.xxx:xxxxx_____")];

	if (portno == -1) goto err_invalid_port;
	    
	snprintf(socket_str, sizeof(socket_str), INET_ADDR_FORMAT ":%u",
		 INET_ADDR_SPLIT(nodeid), portno);

	return socket_str;
	/* error code */
err_invalid_port:
	errno = EINVAL;
	return NULL;
}


int pscom_parse_socket_str(const char *socket_str, int *nodeid, int *portno)
{
	char *lname = NULL;
	char *host;
	char *port = NULL;
	int ret = 0;
	struct sockaddr_in sock;

	if (!socket_str) goto err_arg;
	
	lname = strdup(socket_str);
	if (!lname) goto err_no_mem;
 
	host = strtok_r(lname, ":", &port);
	if (!host) goto err_no_host;
	if (!port) goto err_no_port;

	if (pscom_ascii_to_sockaddr_in(host, port, "tcp", &sock) < 0) goto err_to_sock;

	if (nodeid) *nodeid = (int) ntohl(sock.sin_addr.s_addr);
	if (portno) *portno = (int) ntohs(sock.sin_port);

	if (0) { /* error code */
	err_arg:
	err_no_host:
	err_no_port:
		errno = EINVAL;
	err_no_mem:
	err_to_sock:
		ret = -1;
	}

	if (lname) free(lname);
	return ret;
}


pscom_err_t pscom_connect_socket_str(pscom_connection_t *connection, const char *socket_str)
{
	int nodeid;
	int portno;
	int res;

	res = pscom_parse_socket_str(socket_str, &nodeid, &portno);
	if (res) goto err_parse;

	return pscom_connect(connection, nodeid, portno);
	/* error code */
err_parse:
	if (socket_str) {
		DPRINT(1, "CONNECT (%s) failed : %s", socket_str, pscom_err_str(res));
	} else {
		DPRINT(1, "CONNECT (<null>) failed : %s", pscom_err_str(res));
	}
	return res;
}


const char *
pscom_listen_socket_str(pscom_socket_t *socket)
{
	return pscom_socket_str(pscom_get_nodeid(), socket->listen_portno);
}


const char *
pscom_inetstr(int addr)
{
	static char ret[sizeof("xxx.xxx.xxx.xxx_____")];
	snprintf(ret, sizeof(ret), INET_ADDR_FORMAT,
		 INET_ADDR_SPLIT(addr));
	return ret;
}


const char *
pscom_con_state_str(pscom_con_state_t state)
{
	switch (state) {
	case PSCOM_CON_STATE_NO_RW:	return "norw";
	case PSCOM_CON_STATE_R:		return "ro";
	case PSCOM_CON_STATE_W:		return "wo";
	case PSCOM_CON_STATE_RW:	return "open";
	case PSCOM_CON_STATE_CLOSED:	return "closed";
	}
	
	{
		static char buf[sizeof("state0xXXXXXXXX_____")];
		snprintf(buf, sizeof(buf), "state0x%x", state);
		return buf;
	}
}


const char *
pscom_con_type_str(pscom_con_type_t type)
{
	switch (type) {
	case PSCOM_CON_TYPE_NONE:	return "none";
	case PSCOM_CON_TYPE_LOOP:	return "loop";
	case PSCOM_CON_TYPE_TCP:	return "tcp";
	case PSCOM_CON_TYPE_SHM:	return "shm";
	case PSCOM_CON_TYPE_P4S:	return "p4s";
	case PSCOM_CON_TYPE_GM:		return "gm";
	case PSCOM_CON_TYPE_MVAPI:	return "mvapi";
	case PSCOM_CON_TYPE_OPENIB:	return "openib";
	case PSCOM_CON_TYPE_ELAN:	return "elan";
	case PSCOM_CON_TYPE_DAPL:	return "dapl";
	}
	
	{
		static char buf[sizeof("type0xXXXXXXXX______")];
		snprintf(buf, sizeof(buf), "type0x%x", type);
		return buf;
	}
}


const char *
pscom_con_info_str(pscom_con_info_t *con_info)
{
	static char buf[sizeof("(xxx.xxx.xxx.xxx,pidxxx,0x_xblast_xblast__,XXXXXXXXXXXXXXXX)_____")];
	snprintf(buf, sizeof(buf), "("INET_ADDR_FORMAT",%d,%p,%.8s)",
		 INET_ADDR_SPLIT(con_info->node_id),
		 con_info->pid,
		 con_info->id,
		 con_info->name);
	return buf;
}


const char *pscom_req_state_str(pscom_req_state_t state)
{
	static char buf[sizeof("sendrecvrmarrmaw(Pgpsdec)done_____")];
	const struct names {
		int flag;
		char *name;
	} n[] = {
		{ PSCOM_REQ_STATE_SEND_REQUEST, "send" },
		{ PSCOM_REQ_STATE_RECV_REQUEST, "recv" },

		{ PSCOM_REQ_STATE_RMA_READ_REQUEST, "rmar" },
		{ PSCOM_REQ_STATE_RMA_WRITE_REQUEST, "rmaw" },
		{ PSCOM_REQ_STATE_RENDEZVOUS_REQUEST, "rdvu" },
		
		{ ~0, "(" },
		{ PSCOM_REQ_STATE_PASSIVE_SIDE, "P" },
		{ PSCOM_REQ_STATE_GRECV_REQUEST, "g" },
		{ PSCOM_REQ_STATE_POSTED, "p" },
		{ PSCOM_REQ_STATE_IO_STARTED, "s" },
		{ PSCOM_REQ_STATE_IO_DONE, "d" },
		{ PSCOM_REQ_STATE_ERROR, "e" },
		{ PSCOM_REQ_STATE_CANCELED, "c" },
		{ PSCOM_REQ_STATE_TRUNCATED, "t" },
		{ ~0, ")" },
		{ PSCOM_REQ_STATE_DONE, "done" },
		{ 0, NULL }
	};
	int i;

	buf[0] = 0;
	
	for (i = 0; n[i].flag; i++) {
		if ((state & n[i].flag) || (n[i].flag == ~0)) {
			strcat(buf, n[i].name);
		}
	}
	
	return buf;
}


const char *
pscom_err_str(pscom_err_t error)
{
	static char buf[100];
	switch (error) {
	case PSCOM_SUCCESS:	return "success";
	case PSCOM_ERR_INVALID:	return "Invalid argument";
	case PSCOM_ERR_ALREADY:	return "Operation already in progress";
	case PSCOM_NOT_IMPLEMENTED:	return "Function not implemented";
	case PSCOM_ERR_EOF:	return "End of file";
	case PSCOM_ERR_IOERROR:	return "IO Error";
	case PSCOM_ERR_UNSUPPORTED_VERSION: return "Unsupported version";
	case PSCOM_ERR_STDERROR:
		return strerror(errno);
	}

	snprintf(buf, sizeof(buf), "error %d", error);
	return buf;
}

const char *
pscom_op_str(pscom_op_t operation)
{
	switch (operation) {
	case PSCOM_OP_READ: return "read";
	case PSCOM_OP_WRITE: return "write";
	}
	
	{
		static char buf[sizeof("op%uxxxxx____")];
		snprintf(buf, sizeof(buf), "op%u", operation);
		return buf;
	}
}


const char *pscom_dumpstr(const void *buf, int size)
{
	static char *ret=NULL;
	char *tmp;
	int s;
	char *b;
	if (ret) free(ret);
	ret = (char *)malloc(size * 5 + 4);
	tmp = ret;
	s = size; b = (char *)buf;
	for (; s ; s--, b++){
		tmp += sprintf(tmp, "<%02x>", (unsigned char)*b);
	}
	*tmp++ = '\'';
	s = size; b = (char *)buf;
	for (; s ; s--, b++){
		/* *tmp++ = isprint(*b) ? *b: '.';*/
		*tmp++ = ((*b >= 32) && (*b < 127)) ? *b: '.';
	}
	*tmp++ = '\'';
	*tmp++ = 0;
	return ret;
}
