/*
 *               ParaStation
 *
 * Copyright (C) 2002-2004 ParTec AG, Karlsruhe
 * Copyright (C) 2005-2007 Cluster Competence Center GmbH, Munich
 * All rights reserved.
 *
 * $Id: pscom_ufd.c 5310 2007-10-24 15:26:35Z hauke $
 *
 * pscom_ufd.c: File handling
 *
 * @author
 *         Jens Hauke <hauke@par-tec.de>
 *
 * @file
 ***********************************************************/
#include "pscom_ufd.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* Somewhere in the middle of the GCC 2.96 development cycle, we implemented
   a mechanism by which the user can annotate likely branch directions and
   expect the blocks to be reordered appropriately.  Define __builtin_expect
   to nothing for earlier compilers.  */
#if __GNUC__ == 2 && __GNUC_MINOR__ < 96
#define __builtin_expect(x, expected_value) (x)
#endif

#define likely(x)	__builtin_expect((x),1)
#define unlikely(x)	__builtin_expect((x),0)


void ufd_init(ufd_t *ufd)
{
	memset(ufd, 0, sizeof(*ufd));
	/* ufd->nufds = 0; */
}

int ufd_add(ufd_t *ufd, int fd,
	    void (*can_read)(ufd_t *ufd, ufd_funcinfo_t *ufd_info),
	    void (*can_write)(ufd_t *ufd, ufd_funcinfo_t *ufd_info),
	    int (*poll)(ufd_t *ufd, ufd_funcinfo_t *ufd_info, int timeout),
	    int *pindex,
	    void *priv)
{
	int idx = ufd->nufds;
	struct pollfd *u = &ufd->ufds[idx];
	ufd_funcinfo_t *ui = &ufd->ufds_info[idx];
	
	ufd->nufds++;
	
	if (ufd->nufds > PSCOM_MAX_UFDS) goto error;
	
	u->fd = fd;
	u->events = 0;
	u->revents = 0;
	ui->can_read = can_read;
	ui->can_write = can_write;
	ui->poll = poll;
	ui->pindex = pindex;
	ui->priv = priv;
	
	if (ui->pindex)
		*ui->pindex = idx;
	return idx;
	/* --- */
error:
	fprintf(stderr, "%s(): assert(ufd->nufds > PSCOM_MAX_UFDS) failed!\n", __func__);
	exit(1);
}

void ufd_del(ufd_t *ufd, int fd)
{
	int idx;
	int i;
	
	for (idx = 0; idx < ufd->nufds; idx++) {
		if (ufd->ufds[idx].fd == fd) {
			ufd->nufds--;
			/* move list down. */
			for (i = idx; i < ufd->nufds; i++) {
				ufd->ufds[i] = ufd->ufds[i + 1];
				ufd->ufds_info[i] = ufd->ufds_info[i + 1];
				if (ufd->ufds_info[i].pindex) /* index update */
					*ufd->ufds_info[i].pindex = i;
			}
			break;
		}
	}
	
	return;
}

void ufd_event_set(ufd_t *ufd, int idx, int event)
{
	ufd->ufds[idx].events |= event;
}

void ufd_event_clr(ufd_t *ufd, int idx, int event)
{
	ufd->ufds[idx].events &= ~event;
}

int ufd_poll(ufd_t *ufd, int timeout)
{
	int nfds;
	int i;
	
//    printf("Poll with timeout %d\n", timeout);
	
	if (unlikely(!ufd->nufds)) {
		if (timeout == 0) return 0;
		if (timeout < 0) {
			static int warn = 0;
			if (!warn) {
				fprintf(stderr, "pscom_ufd.c:%d: deadlock detected!\n", __LINE__);
				fflush(stderr);
				warn = 60 * 10 / 10; // warn every 10 min
			}
			warn --;
			timeout = 10 * 1000;
		}
	}
	
	if ((ufd->nufds == 1) &&
	    (ufd->ufds_info[0].poll)) {
		int ret;
		/* Just wait for one event (Maybe one blocking receive) */
		ret = ufd->ufds_info[0].poll(ufd, ufd->ufds_info, timeout);
		if (ret) return 1;
		/* fallback to poll */
	}
	
	nfds = poll(ufd->ufds, ufd->nufds, timeout);
	
	if (nfds <= 0) return 0;
	
	for (i = 0; i < ufd->nufds; i++) {
		struct pollfd *ufds = ufd->ufds + i;
		ufd_funcinfo_t *ufdi = ufd->ufds_info + i;
		if (ufds->revents & POLLIN) {
			ufds->revents &= ~POLLIN;
			ufdi->can_read(ufd, ufdi);
			
			if (ufds->revents & POLLOUT) {
				ufds->revents &= ~POLLOUT;
				ufdi->can_write(ufd, ufdi);
			}
			if (!(--nfds)) return 1;
		} else if (ufds->revents & POLLOUT) {
			ufds->revents &= ~POLLOUT;
			ufdi->can_write(ufd, ufdi);
			if (!(--nfds)) return 1;
		}
	}
	return 1;
}

