/*
 *               ParaStation
 *
 * Copyright (C) 2002-2004 ParTec AG, Karlsruhe
 * Copyright (C) 2005-2007 Cluster Competence Center GmbH, Munich
 *
 * $Id: psoib.c 5489 2008-01-08 11:32:38Z hauke $
 *
 * psoib.c: OPENIB/Infiniband communication
 *
 * @author
 *         Jens Hauke <hauke@par-tec.de>
 *
 * @file
 ***********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <unistd.h>
#include <assert.h>
#include <sys/resource.h> // getrlimit

#include <sysfs/libsysfs.h>
#include <infiniband/verbs.h>

#include "pscom_util.h"
#include "psoib.h"

/* How many connections are allowed ? */
//#define MAX_QP_N	256
// ToDo: Use old values!!!
#define MAX_QP_N	128
/* Size of the send and receive queue */
#define SIZE_SR_QUEUE	16

#define MAX_PENDING_TOKS (SIZE_SR_QUEUE - 6)

/* Completion queue size */
#define SIZE_CQ		(MAX_QP_N * SIZE_SR_QUEUE)
/* MTU on infiniband */
#define IB_MTU_SPEC	IBV_MTU_1024
#define IB_MTU	(16*1024) /* must be < 65536, or change sizeof psoib_msgheader_t.payload,
			     and should be a power of IB_MTU_SPEC */

#define IB_MTU_PAYLOAD	(IB_MTU - sizeof(psoib_msgheader_t))
#define IB_MAX_INLINE	64

/* I got the error
   THH(4): THHUL_qpm_post_send_req: Send queue is full (128 requests outstanding).
   if SEND_NOTIFICATION is disabled */
#define ENABLE_SEND_NOTIFICATION 1


struct hca_info {
    struct ibv_context *ctx;
    struct ibv_cq      *cq; /* handle to cq */
    struct ibv_pd      *pd; /* Protection domain */
};


struct port_info {
    unsigned int port_num;
    uint16_t	 lid;
    hca_info_t *hca_info;
//    VAPI_hca_port_t hca_port;
//    IB_port_t hca_port_idx;
};


typedef struct {
    void *ptr;
    struct ibv_mr *mr;
//    VAPI_mr_hndl_t mr_hndl;
//    VAPI_mr_t rep_mrw; /* responded memory region properties. */
} mem_info_t;


/* Openib specific information about one connection */
struct psoib_con_info {
    /* low level */
    struct ibv_qp *qp;
    struct ibv_context *ctx; // <- copy from hca_info_t
    port_info_t *port_info;
// old:
//    VAPI_hca_hndl_t hca_hndl;
//    VAPI_qp_hndl_t qp_hndl;    
//    VAPI_qp_prop_t qp_prop; /* QP properties */
    uint16_t	lid; /* Base IB_LID. */

    /* send */
    mem_info_t	send_bufs;
    int		send_pos;

    void	*remote_ptr;
    uint32_t	remote_rkey;
    
    /* recv */
    mem_info_t	recv_bufs;
    int		recv_pos;
    
    /* higher level */
    int	n_send_toks;
    int n_recv_toks;
    int n_tosend_toks;

    int con_broken;
};


typedef struct {
    uint16_t	token;
    uint16_t	payload;
    volatile uint32_t	magic;
} psoib_msgheader_t;

#define PSOIB_MAGIC_UNUSED	0
#define PSOIB_MAGIC_IO		1
#define PSOIB_MAGIC_EOF		2

typedef struct {
    char __data[IB_MTU_PAYLOAD];
    psoib_msgheader_t tail;
} psoib_msg_t;

#define PSOIB_LEN(len) ((len + 7) & ~7)
#define PSOIB_DATA(buf, psoiblen) ((char*)(&(buf)->tail) - psoiblen)

/*
 * static variables
 */

static hca_info_t  default_hca;
static port_info_t default_port;

static char *psoib_err_str = NULL;

int psoib_debug = 2;

#define psoib_dprint(level,fmt,arg... ) do {				\
        if ((level) <= psoib_debug) {					\
		fprintf(stderr, "<psoib%5d:"fmt">\n",getpid() ,##arg);	\
		fflush(stderr);						\
	}								\
}while(0);


static
void psoib_err(char *str)
{
    if (psoib_err_str) free(psoib_err_str);

    psoib_err_str = str ? strdup(str) : strdup("");
    return;
}

static
void psoib_err_errno(char *str, int err_no)
{
    const char *vapi_err = strerror(err_no);
    int len = strlen(str) + strlen(vapi_err) + 20;
    char *msg = malloc(len);

    assert(msg);
    
    strcpy(msg, str);
    strcat(msg, " : ");
    strcat(msg, vapi_err);

    psoib_err(msg);
    free(msg);
}

#if 0
/*
 *
 */
void print_hca_cap(VAPI_hca_vendor_t *hca_vendor_p, VAPI_hca_cap_t *hca_cap_p)
{
    fprintf(stderr, "=== HCA vendor info ===\n");
    fprintf(stderr, "\tvendor_id = 0x%08x\n", hca_vendor_p->vendor_id);
    fprintf(stderr, "\tvendor_part_id = %d\n", hca_vendor_p->vendor_part_id);
    fprintf(stderr, "\thw_ver = 0x%08x\n", hca_vendor_p->hw_ver);
    
    fprintf(stderr, "=== HCA capabilities ===\n");
    
    fprintf(stderr, "\tmax_num_qp = %d\n", hca_cap_p->max_num_qp);
    fprintf(stderr, "\tmax_qp_ous_wr = %d\n", hca_cap_p->max_qp_ous_wr);
    fprintf(stderr, "\tflags = 0x%08x\n", hca_cap_p->flags);
    fprintf(stderr, "\tmax_num_sg_ent = %d\n", hca_cap_p->max_num_sg_ent);
    fprintf(stderr, "\tmax_num_sg_ent_rd = %d\n", hca_cap_p->max_num_sg_ent_rd);
    fprintf(stderr, "\tmax_num_cq = %d\n", hca_cap_p->max_num_cq);
    fprintf(stderr, "\tmax_num_ent_cq = %d\n", hca_cap_p->max_num_ent_cq);
    fprintf(stderr, "\tmax_num_mr = %d\n", hca_cap_p->max_num_mr);
    fprintf(stderr, "\tmax_mr_size = "U64_FMT"\n", hca_cap_p->max_mr_size);
    fprintf(stderr, "\tmax_pd_num = %d\n", hca_cap_p->max_pd_num);
    fprintf(stderr, "\tpage_size_cap = %d\n", hca_cap_p->page_size_cap);
    fprintf(stderr, "\tphys_port_num = %d\n", hca_cap_p->phys_port_num);
    fprintf(stderr, "\tmax_pkeys = %d\n", hca_cap_p->max_pkeys);
    fprintf(stderr, "\tnode_guid = %02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x\n", hca_cap_p->node_guid[0],
	    hca_cap_p->node_guid[1],  
	    hca_cap_p->node_guid[2],  
	    hca_cap_p->node_guid[3],  
	    hca_cap_p->node_guid[4],  
	    hca_cap_p->node_guid[5],  
	    hca_cap_p->node_guid[6],  
	    hca_cap_p->node_guid[7]);
    fprintf(stderr, "\tlocal_ca_ack_delay (Log2 4.096usec Max. RX to ACK or NAK delay) = %d\n", hca_cap_p->local_ca_ack_delay);
    fprintf(stderr, "\tmax_qp_ous_rd_atom = %d\n", hca_cap_p->max_qp_ous_rd_atom);
    fprintf(stderr, "\tmax_ee_ous_rd_atom = %d\n", hca_cap_p->max_ee_ous_rd_atom);
    fprintf(stderr, "\tmax_res_rd_atom = %d\n", hca_cap_p->max_res_rd_atom);
    fprintf(stderr, "\tmax_qp_init_rd_atom = %d\n", hca_cap_p->max_qp_init_rd_atom);
    fprintf(stderr, "\tmax_ee_init_rd_atom = %d\n", hca_cap_p->max_ee_init_rd_atom);
    {
	char s[50];
	switch ( hca_cap_p->atomic_cap) {
	case VAPI_ATOMIC_CAP_NONE:
	    strcpy(s, "VAPI_ATOMIC_CAP_NONE");
	    break;
	case VAPI_ATOMIC_CAP_HCA:
	    strcpy(s, "VAPI_ATOMIC_CAP_HCA");
	    break;
	case VAPI_ATOMIC_CAP_GLOB:
	    strcpy(s, "VAPI_ATOMIC_CAP_GLOB");
	    break;
	default:
	    strcpy(s, "invalid capability");
	}
	fprintf(stderr, "\tatomic_cap = %s\n", s);
    }
    fprintf(stderr, "\tmax_ee_num = %d\n", hca_cap_p->max_ee_num);
    fprintf(stderr, "\tmax_rdd_num = %d\n", hca_cap_p->max_rdd_num);
    fprintf(stderr, "\tmax_mw_num = %d\n", hca_cap_p->max_mw_num);
    fprintf(stderr, "\tmax_raw_ipv6_qp = %d\n", hca_cap_p->max_raw_ipv6_qp);
    fprintf(stderr, "\tmax_raw_ethy_qp = %d\n", hca_cap_p->max_raw_ethy_qp);
    fprintf(stderr, "\tmax_mcast_grp_num = %d\n", hca_cap_p->max_mcast_grp_num);
    fprintf(stderr, "\tmax_mcast_qp_attach_num = %d\n", hca_cap_p->max_mcast_qp_attach_num);
    fprintf(stderr, "\tmax_total_mcast_qp_attach_num = %d\n", hca_cap_p->max_total_mcast_qp_attach_num);
    fprintf(stderr, "\tmax_ah_num = %d\n", hca_cap_p->max_ah_num);
}

/*
 *  print_qp_props
 */
void print_qp_props(VAPI_qp_prop_t *qp_props)
{
    fprintf(stderr, "=== QP properties ===\n");
    fprintf(stderr, "\tqp_num = 0x%06x\n", qp_props->qp_num);
    fprintf(stderr, "\tmax_oust_wr_sq = %d\n", qp_props->cap.max_oust_wr_sq);
    fprintf(stderr, "\tmax_oust_wr_rq = %d\n", qp_props->cap.max_oust_wr_rq);
    fprintf(stderr, "\tmax_sg_size_sq = %d\n", qp_props->cap.max_sg_size_sq);
    fprintf(stderr, "\tmax_sg_size_rq = %d\n", qp_props->cap.max_sg_size_rq);
}

#endif

static
struct ibv_device *psoib_get_dev_by_hca_name(const char *in_hca_name)
{
#if 1
    /* new method with ibv_get_device_list() */
    struct ibv_device **dev_list;
    struct ibv_device *ib_dev;
    int dev_list_count;
    
    dev_list = ibv_get_device_list(&dev_list_count);
    if (!dev_list) goto err_no_dev;
    if (!in_hca_name) {
	const char *tmp;
	ib_dev = dev_list[0];

	tmp = ibv_get_device_name(ib_dev);
	
	psoib_dprint(2, "Got IB device \"%s\"", tmp);
	    
	if (!ib_dev) goto err_no_dev2;
    } else {
	int i;
        for (i = 0; i < dev_list_count; i++) {
	    ib_dev = dev_list[i];
	    if (!ib_dev) break;
	    const char *tmp = ibv_get_device_name(ib_dev);
	    if (!strcmp(tmp, in_hca_name)) {
		psoib_dprint(2, "Got IB device \"%s\"", tmp);
		break;
	    } else {
		psoib_dprint(3, "Skip IB device \"%s\" (searching for \"%s\")", tmp, in_hca_name);
	    }
	}
	if (!ib_dev) goto err_no_dev_name;
    }
    ibv_free_device_list(dev_list);

    return ib_dev;
    /* --- */
 err_no_dev:
    psoib_err_errno("ibv_get_devices() failed : No IB dev found", errno);
    return 0;
    /* --- */
 err_no_dev2:
    psoib_err_errno("ibv_get_devices() failed : IB dev list empty", errno);
    ibv_free_device_list(dev_list);
    return 0;
    /* --- */
 err_no_dev_name:
    {
	static char err_str[100];
	snprintf(err_str, 100, "IB device \"%s\" not found", in_hca_name);
	psoib_err_errno(err_str, errno);
	ibv_free_device_list(dev_list);
	return 0;
    }
#else
    /* old method with ibv_get_devices */
    struct dlist *dev_list;
    struct ibv_device *ib_dev;
    
    dev_list = ibv_get_devices();
    
    if (!dev_list) goto err_no_dev;
    
    dlist_start(dev_list);
    if (!in_hca_name) {
	const char *tmp;
	ib_dev = dlist_next(dev_list);

	tmp = ibv_get_device_name(ib_dev);
	psoib_dprint(2, "Got IB device \"%s\"", tmp);
	    
	if (!ib_dev) goto err_no_dev2;
    } else {
	dlist_for_each_data(dev_list, ib_dev, struct ibv_device) {
	    const char *tmp = ibv_get_device_name(ib_dev);
	    if (!strcmp(tmp, in_hca_name)) {
		psoib_dprint(2, "Got IB device \"%s\"", tmp);
		break;
	    } else {
		psoib_dprint(3, "Skip IB device \"%s\" (searching for \"%s\")", tmp, in_hca_name);
	    }
	}
	if (!ib_dev) goto err_no_dev_name;
    }

    return ib_dev;
    /* --- */
 err_no_dev:
    psoib_err_errno("ibv_get_devices() failed : No IB dev found", errno);
    return 0;
    /* --- */
 err_no_dev2:
    psoib_err_errno("ibv_get_devices() failed : IB dev list empty", errno);
    return 0;
    /* --- */
 err_no_dev_name:
    {
	static char err_str[100];
	snprintf(err_str, 100, "IB device \"%s\" not found", in_hca_name);
	psoib_err_errno(err_str, errno);
	return 0;
    }
#endif
}


/* if hca_name == NULL choose first HCA */
static
struct ibv_context *psoib_open_hca(char *hca_name)
{
    struct ibv_device *ib_dev;
    struct ibv_context *ctx;
    
    ib_dev = psoib_get_dev_by_hca_name(hca_name);
    if (!ib_dev) goto err_no_hca;

    ctx = ibv_open_device(ib_dev);
    if (!ctx) goto err_open_device;

    return ctx;
    /* --- */
 err_open_device:
    psoib_err_errno("ibv_open_device() failed", errno);
    return NULL;
    /* --- */
 err_no_hca:
    return NULL;
}

static
struct ibv_cq *psoib_open_cq(struct ibv_context *ctx, int cqe_num)
{
    /* create completion queue - used for both send and receive queues */
    struct ibv_cq *cq;

    errno = 0;
    cq = ibv_create_cq(ctx, cqe_num, NULL, NULL, 0);

    if (!cq) {
	psoib_err_errno("ibv_create_cq() failed", errno);
    }
    
    return cq;
}

static
struct ibv_pd *psoib_open_pd(struct ibv_context *ctx)
{
    /* allocate a protection domain to be associated with QP */
    struct ibv_pd *pd;

    pd = ibv_alloc_pd(ctx);

    if (!pd) {
	psoib_err_errno("ibv_alloc_pd() failed", errno);
    }
    
    return pd;
}

#if 0 
/* INOUT hca_port_idx == -1: scan for first active hca_port */
static
int psoib_open_hca_port(VAPI_hca_hndl_t hca_hndl, int hca_port_idx,
		        VAPI_hca_port_t *hca_port,
			IB_port_t *hca_port_idx_out)
{
    VAPI_ret_t rc;
    IB_port_t port;
    
    assert(hca_port);
    
    if (hca_port_idx > 0) {
	port = hca_port_idx;
	rc = VAPI_query_hca_port_prop(hca_hndl, port, hca_port);
	if (rc != VAPI_OK) goto err_VAPI_query_hca_port_prop;
	
	if (hca_port->state != PORT_ACTIVE) goto err_linkdown;
    } else {
	port = 1;
	while (1) {
	    rc = VAPI_query_hca_port_prop(hca_hndl, port, hca_port);
	    if (rc == VAPI_OK) {
		if (hca_port->state == PORT_ACTIVE)
		    break;
	    }
	    if (++port > 32) goto err_alllinkdown;
	}
    }

    if (psoib_debug) {
	psoib_dprint(1, "Infiniband HCA Port: %d", port);
    }
    *hca_port_idx_out = port;
    
    return 0;
    /* --- */
 err_alllinkdown:
    psoib_err("HCA all Ports are down!");
    return -1;
    /* --- */
 err_linkdown:
    {
	char msg[100];
	snprintf(msg, 99, "HCA Port %d is down!", port);
	psoib_err(msg);
    }
    return -1;
    /* --- */
 err_VAPI_query_hca_port_prop:
    psoib_err_rc("VAPI_query_hca_port_prop() failed", rc);
    return -1;
}
#endif


static
void psoib_vapi_free(hca_info_t *hca_info, mem_info_t *mem_info)
{
    ibv_dereg_mr(/*hca_info->ctx,*/ mem_info->mr);
    mem_info->mr = NULL;
    free(mem_info->ptr);
    mem_info->ptr = NULL;
}


static
void print_mlock_help(void)
{
    static int called = 0;
    struct rlimit rlim;
	
    if (called) return;
    called = 1;

    psoib_dprint(0, "OPENIB: memlock failed. (Check memlock limit in /etc/security/limits.conf or try 'ulimit -l')");
    
    if (!getrlimit(RLIMIT_MEMLOCK, &rlim)) {
	psoib_dprint(0, "Current RLIMIT_MEMLOCK: soft=%lu byte, hard=%lu byte", rlim.rlim_cur, rlim.rlim_max);
    }
}

static
int psoib_vapi_alloc(hca_info_t *hca_info, int size, enum ibv_access_flags access_perm, mem_info_t *mem_info)
{
    mem_info->mr = NULL;
    
    /* Region for buffers */
    mem_info->ptr = valloc(size);
    if (!mem_info->ptr) goto err_malloc;

//    printf("ibv_reg_mr(pd = %p, ptr = %p, size = %d, access_perm = 0x%x)\n",
//	   hca_info->pd, mem_info->ptr, size, access_perm);

    mem_info->mr = ibv_reg_mr(hca_info->pd, mem_info->ptr, size, access_perm);
    if (!mem_info->mr) goto err_reg_mr;
    
    return 0;
    /* --- */
 err_reg_mr:
    free(mem_info->ptr);
    mem_info->ptr = NULL;
    psoib_err_errno("ibv_reg_mr() failed", errno);
    if (errno == ENOMEM) print_mlock_help();
    return -1;
 err_malloc:
    psoib_err_errno("malloc() failed!", errno);
    return -1;
}


void psoib_con_cleanup(psoib_con_info_t *con_info, hca_info_t *hca_info)
{
    if (!hca_info) hca_info = &default_hca;	
    if (con_info->send_bufs.mr) {
	psoib_vapi_free(hca_info, &con_info->send_bufs);
	con_info->send_bufs.mr = 0;
    }
    if (con_info->recv_bufs.mr) {
	psoib_vapi_free(hca_info, &con_info->recv_bufs);
	con_info->recv_bufs.mr = 0;
    }
    if (con_info->qp) {
	ibv_destroy_qp(con_info->qp);
	con_info->qp = 0;
    }
}

/*
 *  move_to_rtr
 */
static
int move_to_rtr(struct ibv_qp *qp,
		unsigned int port_num,
		uint16_t remote_lid, /* remote peer's LID */
		uint32_t remote_qpn) /* remote peer's QPN */
{
    struct ibv_qp_attr attr = {
	.qp_state		= IBV_QPS_RTR,
	.path_mtu		= IB_MTU_SPEC,
	.dest_qp_num		= remote_qpn,
	.rq_psn 		= 0, /* Packet sequence number */
	.max_dest_rd_atomic	= 1, /* Maximum number of oust. RDMA read/atomic as target */
	.min_rnr_timer		= 12, /* Minimum RNR NAK timer (old = 0) */
	.ah_attr		= {
	    .is_global	= 0, /* old av.grh_flag ? */
	    .dlid	= remote_lid,
	    .sl		= 0,  /* Service level bits ??? */
	    .src_path_bits	= 0,
	    .port_num	= port_num
	}
    };
    if (ibv_modify_qp(qp, &attr,
		      IBV_QP_STATE              |
		      IBV_QP_AV                 |
		      IBV_QP_PATH_MTU           |
		      IBV_QP_DEST_QPN           |
		      IBV_QP_RQ_PSN             |
		      IBV_QP_MAX_DEST_RD_ATOMIC |
		      IBV_QP_MIN_RNR_TIMER))
	goto err_ibv_modify_qp;
    
    return 0;
    /* --- */
 err_ibv_modify_qp:
    psoib_err_errno("ibv_modify_qp() move to RTR failed", errno);
    return -1;
}


/*
 *  move_to_rts
 */
static
int move_to_rts(struct ibv_qp *qp)
{
    struct ibv_qp_attr attr = {
	.qp_state	= IBV_QPS_RTS,
	.timeout 	= 14, /* old = 10 */
	.retry_cnt 	= 7,  /* old = 1 */
	.rnr_retry 	= 7,  /* old = 1 */
	.sq_psn 	= 0,  /* Packet sequence number */
	.max_rd_atomic  = 1,  /* Number of outstanding RDMA rd/atomic ops at destination ?*/
    };
    
    if (ibv_modify_qp(qp, &attr,
		      IBV_QP_STATE              |
		      IBV_QP_TIMEOUT            |
		      IBV_QP_RETRY_CNT          |
		      IBV_QP_RNR_RETRY          |
		      IBV_QP_SQ_PSN             |
		      IBV_QP_MAX_QP_RD_ATOMIC))
	goto err_VAPI_modify_qp;
    
    return 0;
    /* --- */
 err_VAPI_modify_qp:
    psoib_err_errno("ibv_modify_qp() move to RTS failed", errno);
    return -1;
}


int psoib_con_init(psoib_con_info_t *con_info, hca_info_t *hca_info, port_info_t *port_info)
{
    //   VAPI_qp_init_attr_t  qp_init_attr;
//    VAPI_ret_t rc;
//    VAPI_qp_cap_t        qp_cap;
    int i;

    if (!hca_info) hca_info = &default_hca;
    if (!port_info) port_info = &default_port;
    
    con_info->ctx = hca_info->ctx;
    con_info->port_info = port_info;
    con_info->qp = NULL;
    con_info->lid = port_info->lid;

    con_info->send_bufs.mr = NULL;
    con_info->recv_bufs.mr = NULL;
    con_info->con_broken = 0;


    {
	struct ibv_qp_init_attr attr = {
	    .send_cq = hca_info->cq,
	    .recv_cq = hca_info->cq,
	    .cap     = {
		//.max_send_wr  = 128, /* Max outstanding WR on the SQ ??*/
		//.max_recv_wr  = 128, /* Max outstanding WR on the RQ ??*/
		//.max_send_sge = 4,   /* Max scatter/gather descriptor entries on the SQ ??*/
		//.max_recv_sge = 4,   /* Max scatter/gather descriptor entries on the RQ */
		.max_send_wr  = 128, /* Max outstanding WR on the SQ ??*/
		.max_recv_wr  = 128, /* Max outstanding WR on the RQ ??*/
		.max_send_sge = 1,   /* Max scatter/gather descriptor entries on the SQ ??*/
		.max_recv_sge = 1,   /* Max scatter/gather descriptor entries on the RQ */
		.max_inline_data = IB_MAX_INLINE,
	    },
	    .qp_type = IBV_QPT_RC
	};

	con_info->qp = ibv_create_qp(hca_info->pd, &attr);
	if (!con_info->qp) goto err_create_qp;
    }

    {
	struct ibv_qp_attr attr;
	
	attr.qp_state        = IBV_QPS_INIT;
	attr.pkey_index      = 0;
	attr.port_num        = port_info->port_num;
	attr.qp_access_flags = IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_REMOTE_READ;
	//                 ToDo:  == VAPI_EN_REM_WRITE | VAPI_EN_REM_READ ??
	
	if (ibv_modify_qp(con_info->qp, &attr,
			  IBV_QP_STATE              |
			  IBV_QP_PKEY_INDEX         |
			  IBV_QP_PORT               |
			  IBV_QP_ACCESS_FLAGS)) goto err_modify_qp;
    }
    
    

    /*
     *  Memory for send and receive bufs
     */
    
    if (psoib_vapi_alloc(hca_info, IB_MTU * SIZE_SR_QUEUE, 0, &con_info->send_bufs))
	goto err_alloc;
    con_info->send_pos = 0;
    
    if (psoib_vapi_alloc(hca_info, IB_MTU * SIZE_SR_QUEUE,
			 IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE,
			 &con_info->recv_bufs))
	goto err_alloc;
    
    /* Clear all receive magics */
    for (i = 0; i < SIZE_SR_QUEUE; i++) {
	psoib_msg_t *msg = ((psoib_msg_t *)con_info->recv_bufs.ptr) + i;
	msg->tail.magic = PSOIB_MAGIC_UNUSED;
    }
    
    con_info->recv_pos = 0;
    
    return 0;
    /* --- */
 err_alloc:
    psoib_con_cleanup(con_info, hca_info);
    goto return_1;
    /* --- */
 err_modify_qp:
    psoib_err_errno("ibv_modify_qp() failed", errno);
    psoib_con_cleanup(con_info, hca_info);
    goto return_1;
    /* --- */
 err_create_qp:
    psoib_err_errno("ibv_create_qp() failed", errno);
    if (errno == ENOMEM) print_mlock_help();
    goto return_1;
    /* --- */
 return_1:
    psoib_dprint(1, "psoib_con_init failed : %s", psoib_err_str);
    return -1;
}


int psoib_con_connect(psoib_con_info_t *con_info, psoib_info_msg_t *info_msg)
{
    int i;

    con_info->remote_ptr = info_msg->remote_ptr;
    con_info->remote_rkey = info_msg->remote_rkey;

    if (move_to_rtr(con_info->qp, con_info->port_info->port_num,
		    info_msg->lid, info_msg->qp_num))
	    goto err_move_to_rtr;
    
    con_info->n_send_toks = 0; // SIZE_SR_QUEUE;
    con_info->n_recv_toks = SIZE_SR_QUEUE;
    con_info->n_tosend_toks = 0;
    if (move_to_rts(con_info->qp))
	    goto err_move_to_rts;

    /* Fill the receive queue */
    for (i = 0; i < SIZE_SR_QUEUE; i++) {
	/* The next call never send tokens to the
	   remote qp, because n_send_toks is 0 */
	psoib_recvdone(con_info);
    }

    /* Now all recv buffers are posted. The same is true for the
       remote qp. n_tosend_toks == SIZE_SR_QUEUE and n_send_toks == 0.
       We can now exchange the tokens without sending a message: */
    assert(con_info->n_send_toks == 0);
    assert(con_info->n_tosend_toks == SIZE_SR_QUEUE);
    assert(con_info->n_recv_toks == 0);

    con_info->n_send_toks = SIZE_SR_QUEUE;
    con_info->n_tosend_toks = 0;
    
    return 0;
    /* --- */
 err_move_to_rtr:
 err_move_to_rts:
    return -1;
}

static
void psoib_cleanup_hca(hca_info_t *hca_info)
{
    if (hca_info->pd) {
	ibv_dealloc_pd(hca_info->pd);
	hca_info->pd = NULL;
    }
    if (hca_info->cq) {
	ibv_destroy_cq(hca_info->cq);
	hca_info->cq = NULL;
    }
    if (hca_info->ctx) {
	ibv_close_device(hca_info->ctx);
	hca_info->ctx = NULL;
    }
}

static
int psoib_init_hca(hca_info_t *hca_info)
{
    hca_info->ctx = NULL;
    hca_info->cq = NULL;
    hca_info->pd = NULL;

    // ToDo: env for hca devname?
    hca_info->ctx = psoib_open_hca(NULL);
    if (!hca_info->ctx) goto err_hca;

    hca_info->cq = psoib_open_cq(hca_info->ctx, SIZE_CQ);
    if (!hca_info->cq) goto err_cq;

    hca_info->pd = psoib_open_pd(hca_info->ctx);
    if (!hca_info->pd) goto err_pd;

    return 0;
    /* --- */
 err_pd:
 err_cq:
 err_hca:
    psoib_cleanup_hca(hca_info);
    return -1;
}

static
int psoib_init_port(hca_info_t *hca_info, port_info_t *port_info)
{
    port_info->hca_info = hca_info;
/*	
    if (psoib_open_hca_port(hca_info->hca_hndl,
			   -1,
			    ,
			   &port_info->port_num)) goto err_port;

*/
    port_info->port_num = 1; // ToDo: Scan all ports!!!
    port_info->hca_info = hca_info;

    {
	struct ibv_port_attr attr;
	if (ibv_query_port(hca_info->ctx, port_info->port_num, &attr))
	    goto err_query_port;
	
	port_info->lid = attr.lid;
    }
    
    return 0;
    /* --- */
 err_query_port:
    psoib_err_errno("ibv_query_port() failed", errno);
    return -1;
}


int psoib_init(void)
{
    static int init_state = 1;
    if (init_state == 1) {
	if (psoib_init_hca(&default_hca)) goto err_hca;

	if (psoib_init_port(&default_hca, &default_port)) goto err_port;
	init_state = 0;
    }
	
    return init_state; /* 0 = success, -1 = error */
    /* --- */
 err_port:
    psoib_cleanup_hca(&default_hca);
 err_hca:
    init_state = -1;
    psoib_dprint(1, "OPENIB disabled : %s", psoib_err_str);
    return -1;
}

static
int psoib_poll(hca_info_t *hca_info, int blocking);

/* returnvalue like write(), except on error errno is negative return */

/* It's important, that the sending side is aligned to IB_MTU_SPEC,
   else we loose a lot of performance!!! */
static
int _psoib_sendv(psoib_con_info_t *con_info, struct iovec *iov, int size, unsigned int magic)
{
    int len;
    int psoiblen;
    psoib_msg_t *_msg;
//    VAPI_sr_desc_t sr_desc;
//    VAPI_sg_lst_entry_t sg_lst;
    int rc;
    psoib_msgheader_t *tail;
    
    if (con_info->con_broken) goto err_broken;

    /* Its allowed to send, if
       At least 2 tokens left or (1 token left AND n_tosend > 0)
    */

    if ((con_info->n_send_toks < 2) &&
	((con_info->n_send_toks < 1) || (con_info->n_tosend_toks == 0))) goto err_busy;

    len = (size <= (int)IB_MTU_PAYLOAD) ? size : (int)IB_MTU_PAYLOAD;
    psoiblen = PSOIB_LEN(len);
    
    _msg = ((psoib_msg_t *)con_info->send_bufs.ptr) + con_info->send_pos;

    tail = (psoib_msgheader_t *)((char*)_msg + psoiblen);
    
    tail->token = con_info->n_tosend_toks;
    tail->payload = len;
    tail->magic = magic;

    /* copy to registerd send buffer */
//    last_msg = _msg;
//    last_iov = iov;
//    last_len = len;
    pscom_memcpy_from_iov((void *)_msg, iov, len);

    {
    	struct ibv_sge list = {
	    .addr 	= (uintptr_t) _msg,
	    .length = psoiblen + sizeof(psoib_msgheader_t),
	    .lkey 	= con_info->send_bufs.mr->lkey,
	};
	struct ibv_send_wr wr = {
	    .next	= NULL,
	    .wr_id	= (uint64_t)con_info,
	    .sg_list	= &list,
	    .num_sge	= 1,
	    .opcode	= IBV_WR_RDMA_WRITE,
	    .send_flags	= (
		(ENABLE_SEND_NOTIFICATION ? IBV_SEND_SIGNALED : 0) | /* no cq entry, if unsignaled */
		((list.length <= IB_MAX_INLINE) ? IBV_SEND_INLINE : 0)),
	    .imm_data	= 42117,

	    .wr.rdma = {
		.remote_addr = (uint64_t)PSOIB_DATA(
		    (((psoib_msg_t *)con_info->remote_ptr) + con_info->send_pos), psoiblen),
		.rkey = con_info->remote_rkey,
	    },
	};

	struct ibv_send_wr *bad_wr;

	rc = ibv_post_send(con_info->qp, &wr, &bad_wr);
    }

    if (rc != 0) goto err_ibv_post_send;

    pscom_forward_iov(iov, len);
    
    con_info->n_tosend_toks = 0;
    con_info->send_pos = (con_info->send_pos + 1) % SIZE_SR_QUEUE;
    con_info->n_send_toks--;

    psoib_poll(&default_hca, 0);

    return len;
    /* --- */
 err_busy:
    psoib_poll(&default_hca, 0);
    return -EAGAIN;
    /* --- */
 err_ibv_post_send:
    if (errno == EAGAIN /* Too many posted work requests ? */) {
	psoib_poll(&default_hca, 0);
//	printf("return2 busy , len = %d\n", size);
	return -EAGAIN;
    } else {
	psoib_err_errno("ibv_post_send() failed", errno);
//	printf("%s\n",psoib_err_str);
	con_info->con_broken = 1;
	return -EPIPE;
    }
    /* --- */
 err_broken:
    return -EPIPE;
}


int psoib_sendv(psoib_con_info_t *con_info, struct iovec *iov, int size)
{
	return _psoib_sendv(con_info, iov, size, PSOIB_MAGIC_IO);
}


void psoib_send_eof(psoib_con_info_t *con_info)
{
	_psoib_sendv(con_info, NULL, 0, PSOIB_MAGIC_EOF);
	con_info->con_broken = 1; // Do not send more
}

	
void psoib_recvdone(psoib_con_info_t *con_info)
{
    psoib_msg_t *msg;
    
    msg = ((psoib_msg_t *)con_info->recv_bufs.ptr) + con_info->recv_pos;
    
    con_info->n_tosend_toks++;
    con_info->n_recv_toks--;
    con_info->recv_pos = (con_info->recv_pos + 1) % SIZE_SR_QUEUE;

    if (con_info->n_tosend_toks >= MAX_PENDING_TOKS) {
	//while (psoib_sendv(con_info, NULL, 0) == -EAGAIN);
	psoib_sendv(con_info, NULL, 0);
    }
}


/* returnvalue like read() , except on error errno is negative return */
int psoib_recvlook(psoib_con_info_t *con_info, void **buf)
{
#if 1 // Simpler loop because:
	// assert(con_info->n_recv_toks == 0) as long as we only poll!
	while (1) {
		psoib_msg_t *msg =
			((psoib_msg_t *)con_info->recv_bufs.ptr) + con_info->recv_pos;

		unsigned int magic = msg->tail.magic;

		if (!magic) { // Nothing received
			*buf = NULL;
			return (con_info->con_broken) ? -EPIPE : -EAGAIN;
		}
		
		msg->tail.magic = PSOIB_MAGIC_UNUSED;

		/* Fresh tokens ? */
		con_info->n_send_toks += msg->tail.token;
		con_info->n_recv_toks++;

		unsigned int len = msg->tail.payload;

		*buf = PSOIB_DATA(msg, PSOIB_LEN(len));
		if (len || (magic == PSOIB_MAGIC_EOF)) {
			// receive data or EOF
			return len;
		}

		/* skip 0 payload packages (probably fresh tokens) */
		psoib_recvdone(con_info);
	}
#else
    unsigned int magic;
    /* Check for new packages */
    {
	psoib_con_info_t *con = con_info;
	psoib_msg_t *msg = ((psoib_msg_t *)con->recv_bufs.ptr) +
	    ((con->recv_pos + con->n_recv_toks) % SIZE_SR_QUEUE);
	magic = msg->tail.magic;
	
	if (magic) {
//	    printf("receive magic %08x\n", msg->tail.magic);
	    msg->tail.magic = PSOIB_MAGIC_UNUSED;

	    /* Fresh tokens ? */
	    con->n_send_toks += msg->tail.token;
	    con->n_recv_toks++;
	}
    }

    while (con_info->n_recv_toks > 0) {
	psoib_msg_t *msg = ((psoib_msg_t *)con_info->recv_bufs.ptr) + con_info->recv_pos;
	int len = msg->tail.payload;
	
	*buf = PSOIB_DATA(msg, PSOIB_LEN(len));
	if (len || (magic == PSOIB_MAGIC_EOF)) {
	    // ToDo: This could be the wrong magic!!!	
	    return len;
	}
	/* skip 0 payload packages */
	psoib_recvdone(con_info);
    }

    if (con_info->con_broken) {
	return -EPIPE;
    } else {
	return -EAGAIN;
    }
#endif
}

static
int psoib_check_cq(hca_info_t *hca_info)
{
    struct ibv_wc wc;
    int rc;
    
    rc = ibv_poll_cq(hca_info->cq, 1, &wc);

    if (rc == 1) {
	if (wc.opcode == IBV_WC_RDMA_WRITE /* == VAPI_CQE_SQ_RDMA_WRITE ?*/) {
	    /* RDMA write done */
	    psoib_con_info_t *con = (psoib_con_info_t *)(unsigned long)wc.wr_id;
	    if (wc.status == IBV_WC_SUCCESS) {
//		printf("RDMA write done... recv: %d tosend: %d send: %d\n",
//		       con->n_recv_toks, con->n_tosend_toks, con->n_send_toks);
	    } else {
		psoib_dprint(1, "Failed RDMA write request (status %d). Connection broken!", wc.status);
		con->con_broken = 1;
	    }
	} else if (wc.opcode == (IBV_WC_SEND  | IBV_WC_RECV) /* == VAPI_CQE_RQ_SEND_DATA*/) {
	    /* receive something */
	    psoib_con_info_t *con = (psoib_con_info_t *)(unsigned long)wc.wr_id;
	    printf("Recv done... recv: %d tosend: %d send: %d\n",
		   con->n_recv_toks, con->n_tosend_toks, con->n_send_toks);
	    if (wc.status == IBV_WC_SUCCESS) {
		psoib_msg_t *msg;
		msg = ((psoib_msg_t *)con->recv_bufs.ptr) +
		    ((con->recv_pos + con->n_recv_toks) % SIZE_SR_QUEUE);

		/* Fresh tokens ? */
		con->n_send_toks += msg->tail.token;
		con->n_recv_toks++;
	    } else {
		psoib_dprint(1, "Failed receive request (status %d). Connection broken!\n", wc.status);
		con->con_broken = 1;
	    }
	} else if (wc.opcode == IBV_WC_SEND /* VAPI_CQE_SQ_SEND_DATA */) {
	    /* Send done */
	    psoib_con_info_t *con = (psoib_con_info_t *)(unsigned long)wc.wr_id;
	    if (wc.status == IBV_WC_SUCCESS) {
		printf("Send done... recv: %d tosend: %d send: %d\n",
		       con->n_recv_toks, con->n_tosend_toks, con->n_send_toks);
	    } else {
		psoib_dprint(1, "Failed send request (status %d). Connection broken!\n", wc.status);
		con->con_broken = 1;
	    }
	} else {
	    psoib_dprint(0, "Unknown Opcode: %d\n", wc.opcode);
	}
    }   
    return rc;
#if 0
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    VAPI_wc_desc_t comp_desc;
    VAPI_ret_t rc;
    rc = VAPI_poll_cq(hca_info->hca_hndl, hca_info->cq_hndl, &comp_desc);
    if (rc == VAPI_OK) {
	if (comp_desc.opcode == VAPI_CQE_SQ_RDMA_WRITE) {
	    /* RDMA write done */
	    psoib_con_info_t *con = (psoib_con_info_t *)(unsigned long)comp_desc.id;
	    if (comp_desc.status == VAPI_SUCCESS) {
		//printf("RDMA write done... recv: %d tosend: %d send: %d\n",
		//       con->n_recv_toks, con->n_tosend_toks, con->n_send_toks);
	    } else {
		if (psoib_debug > 0) {
		    fprintf(stderr, "Failed RDMA write request (status %d). Connection broken!\n",
			    comp_desc.status);
		}
		con->con_broken = 1;
	    }
	} else if (comp_desc.opcode == VAPI_CQE_RQ_SEND_DATA) {
	    /* receive something */
	    psoib_con_info_t *con = (psoib_con_info_t *)(unsigned long)comp_desc.id;
//	    printf("Recv done... recv: %d tosend: %d send: %d\n",
//		   con->n_recv_toks, con->n_tosend_toks, con->n_send_toks);
	    if (comp_desc.status == VAPI_SUCCESS) {
		psoib_msg_t *msg;
		msg = ((psoib_msg_t *)con->recv_bufs.ptr) +
		    ((con->recv_pos + con->n_recv_toks) % SIZE_SR_QUEUE);

		/* Fresh tokens ? */
		con->n_send_toks += msg->tail.token;
		con->n_recv_toks++;
	    } else {
		if (psoib_debug > 0) {
		    fprintf(stderr, "Failed receive request (status %d). Connection broken!\n",
			    comp_desc.status);
		}
		con->con_broken = 1;
	    }
	} else if (comp_desc.opcode == VAPI_CQE_SQ_SEND_DATA) {
	    /* Send done */
	    psoib_con_info_t *con = (psoib_con_info_t *)(unsigned long)comp_desc.id;
	    if (comp_desc.status == VAPI_SUCCESS) {
//		printf("Send done... recv: %d tosend: %d send: %d\n",
//		       con->n_recv_toks, con->n_tosend_toks, con->n_send_toks);
	    } else {
		if (psoib_debug > 0) {
		    fprintf(stderr, "Failed send request (status %d). Connection broken!\n",
			   comp_desc.status);
		}
		con->con_broken = 1;
	    }
	} else {
	    fprintf(stderr, "Unknown Opcode: %d\n", comp_desc.opcode);
	}
    }   
    return rc;
#endif
}

static
int psoib_poll(hca_info_t *hca_info, int blocking)
{
    int rc;

    do {
	rc = psoib_check_cq(hca_info);
    } while (blocking && (rc != 0/*VAPI_CQ_EMPTY*/));

//    if (psoib_debug &&
//	(rc != VAPI_CQ_EMPTY) &&
//	(rc != VAPI_OK)) {
//	fprintf(stderr, "psoib_poll: %s: %s\n", VAPI_strerror_sym(rc), VAPI_strerror(rc));
//    }

    return (rc == 0 /*VAPI_CQ_EMPTY*/);
}


psoib_con_info_t *psoib_con_create(void)
{
	psoib_con_info_t *con_info = malloc(sizeof(*con_info));
	return con_info;
}


void psoib_con_free(psoib_con_info_t *con_info)
{
	free(con_info);
}


void psoib_con_get_info_msg(psoib_con_info_t *con_info /* in */, psoib_info_msg_t *info_msg /* out */)
{
	info_msg->lid = con_info->lid;
	info_msg->qp_num = con_info->qp->qp_num;
	info_msg->remote_ptr = con_info->recv_bufs.ptr;
	info_msg->remote_rkey = con_info->recv_bufs.mr->rkey;
}
